import logging
import re
from collections import namedtuple
from textwrap import dedent

from bottle import request, HTTPError

from ...errors import NotificationError
from ...notification import NotificationMgmt, Notification
from ...postgres import pg_escape


logger = logging.getLogger(__package__)


class FileSetting(namedtuple('FileSetting', ['name', 'setting', 'sourcefile',
                                             'sourceline'])):
    pass


def get_settings_categories(conn):
    return {'categories': [row['category'] for row in conn.query("""\
    SELECT DISTINCT(category) FROM pg_settings ORDER BY category
    """)]}


def get_setting(conn, name):
    return conn.queryscalar(
        "SELECT setting FROM pg_settings WHERE name = %s", (name,))


def preformat(setting, type):
    if setting.startswith("'") and setting.endswith("'"):
        setting = setting[1:-1]
    if type == 'bool':
        if setting == 'true':
            setting = 'on'
        elif setting == 'false':
            setting = 'off'
    return setting


def format_setting(setting, type, unit=None):
    if not setting:
        return
    if type == 'integer':
        setting = int(human_to_number(setting, unit))
    elif type == 'real':
        setting = float(setting)
    return setting


def get_settings(conn, category=None, search=None):
    where = ''
    if search:
        where = dedent("""\
        WHERE name ILIKE '%{0}%'
         OR short_desc ILIKE '%{0}%'
         OR extra_desc ILIKE '%{0}%'
        """).format(search)
    query = dedent("""\
    SELECT
      name,
      setting,
      current_setting(name) AS current_setting,
      unit,
      vartype,
      min_val, max_val, enumvals,
      context, category,
      short_desc || ' ' || coalesce(extra_desc, '') AS desc,
      boot_val, reset_val,
      pending_restart
    FROM pg_settings
    %s
    ORDER BY category, name
    """) % where

    ret = {}
    for row in conn.query(query):
        if category and category != row['category']:
            continue

        rows = ret.setdefault(row['category'], [])
        enumvals = row['enumvals']
        if enumvals is not None:
            # format enumvals as before switching from tpc to psycopg2
            enumvals = '{%s}' % ','.join(enumvals)
        rows.append({
            'name': row['name'],
            'setting': row['setting'],
            'setting_raw': row['current_setting'],
            'unit': row['unit'],
            'vartype': row['vartype'],
            'min_val': row['min_val'],
            'max_val': row['max_val'],
            'boot_val': row['boot_val'],
            'reset_val': row['reset_val'],
            'enumvals': enumvals,
            'context': row['context'],
            'desc': row['desc'],
            'pending_restart': row['pending_restart'],
        })

    return [{'category': k, 'rows': v} for k, v in ret.items()]


def human_to_number(h_value, h_unit=None, h_type=int):
    units = ['B', 'KB', 'MB', 'GB', 'TB', 'PB', 'EB', 'YB', 'ZB']
    re_unit = re.compile(r'([0-9.]+)\s*([KMGBTPEYZ]?B)$', re.IGNORECASE)
    m_value = re_unit.match(str(h_value))
    factor = 1
    if h_unit:
        m_unit = re_unit.match(str(h_unit))
        if m_unit:
            factor = int(m_unit.group(1))
            h_unit = str(m_unit.group(2))

    if m_value:
        p_num = m_value.group(1)
        p_unit = m_value.group(2)
        m = 0
        for u in units:
            if h_unit and h_unit.lower() == u.lower():
                m = 0
            if u.lower() == p_unit.lower():
                return (int(p_num) * (1024 ** m)) / factor
            else:
                m += 1

    # Valid time units are ms (milliseconds), s (seconds), min (minutes),
    # h (hours), and d (days
    re_unit = re.compile(r'([0-9.]+)\s*(us|ms|s|min|h|d)$')
    m_unit = re_unit.match(str(h_value))
    if h_unit == 'ms':
        mult = {'us': 0.001, 'ms': 1, 's': 1000, 'min': 60000, 'h': 3600000,
                'd': 86400000}
    elif h_unit == 's':
        mult = {'ms': -1000, 's': 1, 'min': 60, 'h': 3600, 'd': 86400}
    elif h_unit == 'min':
        mult = {'ms': -60000, 's': -60, 'min': 1, 'h': 60, 'd': 1440}
    elif h_unit == 'h':
        mult = {'ms': -3600000, 's': -3600, 'min': -60, 'h': 1, 'd': 24}
    elif h_unit == 'd':
        mult = {'ms': -86400000, 's': -86400, 'min': -1440, 'h': -24, 'd': 1}
    else:
        mult = {'ms': 1, 's': 1, 'min': 1, 'h': 1, 'd': 1}

    if m_unit:
        p_num = m_unit.group(1)
        p_unit = m_unit.group(2)
        if mult[p_unit] > 0:
            return (h_type(p_num) * mult[p_unit])
        else:
            return (h_type(p_num) / abs(mult[p_unit]))

    return h_value


def get_settings_status(conn):
    settings = get_settings(conn)
    pending_restart_changes = []
    pending_restart = False
    for category in settings:
        for row in category['rows']:
            if row['pending_restart']:
                pending_restart = True
                pending_restart_changes.append(row)
    return {
        'restart_pending': pending_restart,
        'restart_changes': pending_restart_changes,
    }


def post_settings(app, conn, current, update):
    ret = {'settings': []}
    do_not_check_names = ['unix_socket_permissions', 'log_file_mode']
    for setting in update:
        if 'name' not in setting \
           or 'setting' not in setting:
            raise HTTPError(406, "setting item malformed.")
        checked = False
        try:
            for pg_config_category in current:
                for item in pg_config_category['rows']:
                    if item['name'] == setting['name']:
                        if item['name'] in do_not_check_names:
                            checked = True
                            raise Exception()
                        if item['vartype'] == 'integer':
                            # Integers handling.
                            if item['min_val'] and \
                               item['unit'] and \
                               (int(human_to_number(setting['setting'],
                                    item['unit'])) <
                                   int(item['min_val'])):
                                raise HTTPError(406, "%s: Invalid setting." %
                                                     (item['name']))
                            if item['max_val'] and \
                               item['unit'] and \
                               (int(human_to_number(setting['setting'],
                                    item['unit'])) >
                                   int(item['max_val'])):
                                raise HTTPError(406, "%s: Invalid setting." %
                                                     (item['name']))
                            setting['setting'] = pg_escape(setting['setting'])
                            if ((setting['setting'].startswith("'") and
                                 setting['setting'].endswith("'")) or
                                (setting['setting'].startswith('"') and
                                 setting['setting'].endswith('"'))):
                                setting['setting'] = setting['setting'][1:-1]
                            if setting['setting'] == '':
                                setting['setting'] = None
                            checked = True
                        if item['vartype'] == 'real':
                            setting['setting'] \
                                = human_to_number(setting['setting'],
                                                  item['unit'],
                                                  float)
                            # Real handling.
                            if item['min_val'] and \
                               (float(setting['setting']) <
                                   float(item['min_val'])):
                                raise HTTPError(406, "%s: Invalid setting." %
                                                     (item['name']))
                            if item['max_val'] and \
                               (float(setting['setting']) >
                                   float(item['max_val'])):
                                raise HTTPError(406, "%s: Invalid setting." %
                                                     (item['name']))
                            checked = True
                        if item['vartype'] == 'bool':
                            # Boolean handling.
                            if setting['setting'].lower() not in \
                               ['on', 'off']:
                                raise HTTPError(
                                    406, 'Invalid setting: %s.' %
                                         (setting['setting'].lower()))
                            checked = True
                        if item['vartype'] == 'enum':
                            # Enum handling.
                            if len(item['enumvals']) > 0:
                                enumvals = [
                                    re.sub(r"^[\"\'](.+)[\"\ ']$",
                                           r"\1", enumval)
                                    for enumval
                                    in item['enumvals'][1:-1].split(',')]
                                if ((setting['setting'].startswith("'") and
                                     setting['setting'].endswith("'")) or
                                    (setting['setting'].startswith('"') and
                                     setting['setting'].endswith('"'))):
                                    setting['setting'] = \
                                        setting['setting'][1:-1]
                                if setting['setting'] not in enumvals:
                                    raise HTTPError(
                                        406,
                                        'Invalid setting: %s.' %
                                        (setting['setting']))
                                checked = True
                        if item['vartype'] == 'string':
                            # String handling.
                            # setting must be escaped.
                            setting['setting'] = pg_escape(
                                str(setting['setting']))
                            if ((setting['setting'].startswith("'") and
                                 setting['setting'].endswith("'")) or
                                (setting['setting'].startswith('"') and
                                 setting['setting'].endswith('"'))):
                                setting['setting'] = setting['setting'][1:-1]
                            if setting['setting'] == '':
                                setting['setting'] = None
                            checked = True
                        raise Exception()
        except HTTPError as e:
            raise HTTPError(e.code, e.message['error'])
        except Exception:
            pass
        if not checked:
            raise HTTPError(406, 'Parameter %s can\'t be checked.' %
                                 (setting['name']))
        if 'force' not in setting:
            setting['force'] = 'false'
        if ((item['vartype'] == 'integer' and
            setting['setting'] != item['setting_raw']) or
            (item['vartype'] == 'real' and
            float(setting['setting']) != float(item['setting'])) or
            (item['vartype'] not in ['integer', 'real'] and
            setting['setting'] != item['setting'])) or \
                (setting['force'] == 'true'):
            # At this point, all incoming parameters have been checked.
            if setting['setting']:
                query = "ALTER SYSTEM SET {} TO '{}'".format(
                    setting['name'], setting['setting'])
            else:
                query = "ALTER SYSTEM RESET %s;" % (setting['name'])

            logger.debug(query)

            # Push a notification on setting change.
            try:
                NotificationMgmt.push(
                    app.config,
                    Notification(
                        username=request.username,
                        message="Setting '{}' changed: '{}' -> '{}'".format(
                            item['name'],
                            item['setting_raw'],
                            setting['setting'])))
            except NotificationError as e:
                logger.error(e.message)

            try:
                conn.execute(query)
            except Exception as e:
                raise HTTPError(408, "{}: {}".format(setting['name'], e))
            ret['settings'].append({
                'name': item['name'],
                'setting': setting['setting'],
                'previous_setting': item['setting_raw'],
                'restart': True if item['context'] in
                ['internal', 'postmaster'] else False
            })
    # Reload PG configuration.
    conn.execute("SELECT pg_reload_conf()")
    # Push a notification.
    try:
        NotificationMgmt.push(app.config,
                              Notification(username=request.username,
                                           message="PostgreSQL reload"))
    except NotificationError as e:
        logger.error(e.message)

    return ret



import logging

from bottle import Bottle, default_app, HTTPError

from ...tools import now
from ...toolkit.configuration import OptionSpec


bottle = Bottle()
logger = logging.getLogger(__name__)


query = """\
SELECT
  rolname,
  datname,
  pgss.*
FROM pg_stat_statements pgss
JOIN pg_authid ON pgss.userid = pg_authid.oid
JOIN pg_database ON pgss.dbid = pg_database.oid
"""


@bottle.get("/")
def get_statements(pgpool):
    """Return a snapshot of latest statistics of executed SQL statements
    """
    app = default_app().temboard
    config = app.config
    dbname = config.statements.dbname
    snapshot_datetime = now()
    try:
        for connect in pgpool.auto_reconnect():
            with connect(dbname) as conn:
                data = list(conn.query(query))
    except Exception as e:
        discover = app.discover.ensure_latest()
        pg_version = discover['postgres']['version_num']
        if (
            pg_version < 90600 or
            'relation "pg_stat_statements" does not exist' in str(e)
        ):
            raise HTTPError(
                404, "pg_stat_statements not enabled on database %s" % dbname
            )
        logger.error(
            "Failed to get pg_stat_statements data on database %s: %s",
            dbname,
            e,
        )
        raise HTTPError(500, e)
    else:
        return {"snapshot_datetime": snapshot_datetime, "data": data}


class StatementsPlugin:
    PG_MIN_VERSION = (90500, 9.5)
    s = "statements"
    option_specs = [OptionSpec(s, "dbname", default="postgres")]
    del s

    def __init__(self, app, **kw):
        self.app = app
        self.app.config.add_specs(self.option_specs)

    def load(self):
        default_app().mount('/statements', bottle)

    def unload(self):
        self.app.config.remove_specs(self.option_specs)



import os
import logging
import socket
import sys
from getpass import getpass
from textwrap import dedent
from sys import stdout
from urllib.error import HTTPError
from urllib.parse import urlparse

from ..errors import UserError
from ..toolkit.app import SubCommand
from ..toolkit.http import TemboardClient
from ..tools import validate_parameters
from .app import app


try:
    input = raw_input
except NameError:
    pass

logger = logging.getLogger(__name__)
T_USERNAME = b'(^[a-z0-9]{3,16}$)'
T_PASSWORD = b'(^.{5,32}$)'


@app.command
class Register(SubCommand):
    """Register a PostgreSQL instance to a temBoard UI."""

    def define_arguments(self, parser):
        parser.description = dedent("""\

        This command is interactive. register will ask you username and
        password to call temBoard UI API.

        """)

        parser.add_argument(
            '--host',
            dest='host',
            help="Agent address accessible from UI. Default: %(default)s",
            default=socket.getfqdn(),
        )
        self.app.config_specs['temboard_port'].add_argument(
            parser, '-p', '--port',
            help="Agent listening TCP port. Default: %(default)s",
        )
        parser.add_argument(
            '-g', '--groups',
            dest='groups',
            help="Instance groups list, comma separated. Default: %(default)s",
            default=None,
        )
        parser.add_argument(
            '--ui-url',
            dest='temboard_ui_url',
            metavar='URL',
            help="temBoard UI address to register to.",
        )
        super(Register, self).define_arguments(parser)

    def main(self, args):
        agent_hostport = "%s:%s" % (args.host, self.app.config.temboard.port)
        agent_baseurl = "https://%s" % agent_hostport

        logger.info("Working for agent listening at %s.", agent_baseurl)

        agentclient = TemboardClient.factory(
            self.app.config,
            args.host, self.app.config.temboard.port,
        )
        ui_url_raw = self.app.config.temboard.ui_url.rstrip('/')
        ui_url = urlparse(ui_url_raw)
        uiclient = TemboardClient.factory(
            self.app.config,
            scheme=ui_url.scheme, host=ui_url.hostname, port=ui_url.port,
        )

        try:
            # Getting system/instance informations using agent's discovering
            # API.
            logger.info("Discovering system and PostgreSQL ...")
            response = agentclient.get('/discover')
            response.raise_for_status()
            discover = response.json()
            logger.info(
                "temBoard agent for %s instance serving %s on port %s.",
                discover['postgres']['version_summary'],
                discover['postgres']['data_directory'],
                discover['postgres']['port'],
            )
        except OSError as e:
            logger.error("Failed to connect to agent: %s", e)
            logger.error("Is agent %s running?", agent_hostport)
            raise UserError("Connection failure.")
        except TemboardClient.Error as e:
            raise UserError(str(e))

        groups = args.groups
        if groups:
            groups = args.groups.split(',')
        else:
            logger.warning("Instance without groups are hidden in UI!")

        try:
            logger.info("Verifying signing key.")
            response = uiclient.get(ui_url.path + '/signing.key')
            response.raise_for_status()
            wanted = response.read()
            with open(self.app.config.temboard.signing_public_key, 'rb') as fo:
                configured = fo.read()

            if wanted != configured:
                logger.error(
                    "Agent %s is not configured with UI %s signing key.",
                    agent_hostport, ui_url.netloc,
                )
                logger.error(
                    "Use temboard-agent fetch-key --force "
                    "to accept %s signing key.",
                    ui_url.netloc,
                )
                raise UserError("Signature mismatch.")
        except OSError as e:
            logger.error("Failed to connect to UI: %s", e)
            logger.error("Is UI %s running?", ui_url.netloc)
            raise UserError("Connection failure.")
        except TemboardClient.Error as e:
            raise UserError("UI returned error: %s" % e)

        try:
            logger.info("Login at %s ...", ui_url_raw)
            username = ask_username()
            password = ask_password()
        except EOFError:
            # Write new line after prompt.
            sys.stderr.write("\n")
            raise UserError("Exiting.")

        try:
            response = uiclient.post(
                ui_url.path + '/json/login',
                {'username': username, 'password': password},
            )
            response.raise_for_status()

            # POSTing new instance
            logger.info("Registering instance/agent in %s ...", ui_url_raw)
            path = ui_url.path + '/json/register/instance'
            response = uiclient.post(path, {
                'agent_key': app.config.temboard['key'],
                'agent_address': args.host,
                'agent_port': str(app.config.temboard['port']),
                'discover': discover,
                # Loosely reuse agent plugin as UI plugins.
                'plugins': discover['temboard']['plugins'],
                'groups': groups
            })
            response.raise_for_status()
            dashboard_url = ui_url_raw + '/server/%s/%s/dashboard' % (
                args.host, self.app.config.temboard.port,
            )
            logger.info(
                "Instance registered. Manage it at %s.", dashboard_url)
        except TemboardClient.Error as e:
            raise UserError("UI returned error: %s" % e)

        return 0


def ask_password():
    try:
        raw_pass = os.environ['TEMBOARD_UI_PASSWORD']
    except KeyError:
        raw_pass = getpass(" Password: ")

    try:
        password = raw_pass
        validate_parameters({'password': password},
                            [('password', T_PASSWORD, False)])
    except HTTPError:
        stdout.write("Invalid password.\n")
        return ask_password()
    return password


def ask_username():
    try:
        raw_username = os.environ['TEMBOARD_UI_USER']
    except KeyError:
        raw_username = input(" Username: ")

    try:
        username = raw_username
        validate_parameters({'username': username},
                            [('username', T_USERNAME, False)])
    except HTTPError:
        stdout.write("Invalid username.\n")
        return ask_username()
    return username



import sys

from bottle import default_app

from .cli.app import app
from .web.app import create_app


def main():
    default_app.pop()  # Remove default app.
    default_app.push(create_app(app))

    # Import core HTTP routes
    __import__(__package__ + '.web.core')

    # Import commands
    __import__(__package__ + '.cli.discover')
    __import__(__package__ + '.cli.fetch_key')
    __import__(__package__ + '.cli.register')
    __import__(__package__ + '.cli.routes')
    __import__(__package__ + '.cli.runscript')
    __import__(__package__ + '.cli.serve')
    __import__(__package__ + '.cli.tasks')
    __import__(__package__ + '.cli.web')
    return app()


if '__main__' == __name__:
    sys.exit(main())



import logging
import re
import os
import time
import json
from contextlib import closing

import psycopg2
from psycopg2.extensions import parse_dsn
from psycopg2.extras import PhysicalReplicationConnection

from ...inventory import SysInfo
from ...plugins.maintenance.functions import INDEX_BTREE_BLOAT_SQL
from ...toolkit.utils import utcnow

from . import db


logger = logging.getLogger(__package__)


def load_probes(options, home):
    """Give a list of probe objects, ready to run."""

    # All probes classes names start with "probe_", search for
    # classes and get an object
    probes = []
    r = re.compile(r'^probe_(\w+)$')
    for c in globals().keys():
        m = r.search(c)
        if m is not None \
           and issubclass(globals()[c], Probe) \
           and (m.group(1) in options['probes'] or '*' in options['probes']):
            o = eval(c + "(options)")
            o.set_home(home)
            probes.append(o)
            logger.debug("Loaded probe: %s.", o.get_name())

    return probes


def run_probes(probes, pool, instances, delta=True):
    """Execute the probes."""

    now = utcnow()
    logger.info("Running probes at %s.", now.isoformat())
    # Output is a mapping of probe names with lists. Each probe returns
    # a list of dicts(metric -> value).
    output = {}

    for p in probes:
        out = []
        if delta is False:
            p.delta_key = None
            p.delta_columns = None

        if p.level == 'host':
            if not p.check():
                continue
            logger.info("Running host probe %s.", p.get_name())
            try:
                out = p.run()
            except Exception as e:
                logger.error("Probe failure: %s", e)
                continue

        else:
            if p.level not in ('instance', 'database'):
                raise Exception("Unknown probe level: %s", p.level)

            i, = instances  # We are now mono-instance
            if not i['available']:
                continue

            if not p.check(i['version_num']):
                logger.warning(
                    "Unsupported PostgreSQL version for probe %s.",
                    p.get_name())
                continue

            logger.info("Running %s probe %s.", p.level, p.get_name())
            if p.level == 'instance':
                dbnames = [i['database']]
            else:
                dbnames = [db['dbname'] for db in i['dbnames']]

            for dbname in dbnames:
                conninfo = dict(i, dbname=dbname)
                try:
                    out += p.run(pool.getconn(dbname=dbname), conninfo)
                except Exception as e:
                    logger.error("Probe failure: %s", e)
                    raise
                    continue

        for record in out:
            record['datetime'] = now
        output[p.get_name()] = out

    logger.info("Finished probes run.")
    return output


def parse_primary_conninfo(pci):
    # Parse primary_conninfo string picked up from recovery.conf file
    m = re.match(r'.*primary_conninfo\s*=\s*\'(.*)\'[^\']*$', pci)
    if not m:
        raise Exception("Unable to parse primary_conninfo.")

    # Unquote string. When coming from recovery.conf, single quotes can be
    # doubled or escaped with a backslash
    c = m.group(1).replace("''", "'")
    c = c.replace("\\'", "'")

    # Keep only compat parameters because we may be bound to an old libpq.
    well_known_params = [
        p for p in c.split() if (
            p.startswith('user') or
            p.startswith('password') or
            p.startswith('host') or
            p.startswith('port') or
            p.startswith('sslmode')
        )
    ]

    c = ' '.join(well_known_params)

    return c


def get_primary_conninfo(conn):
    # Read and parse primary_conninfo from recovery.conf file if any or from
    # pg_settings if PG version is upper or equal to 12.
    # It returns a tuple: (host, port, user, password)

    # Check if we are in recovery mode *and* recovery file exists
    if conn.server_version < 120000:
        # Note: won't work with PG12
        query = """\
        SELECT (pg_is_in_recovery() AND (
          length(
            coalesce(
              pg_stat_file('recovery.conf', true)::TEXT, ''::TEXT
            )
          ) > 0
        )) AS is_in_recovery
        """
    else:
        query = """\
        SELECT (pg_is_in_recovery() AND (
            length(
            coalesce(
                pg_stat_file('standby.signal', true)::TEXT, ''::TEXT
            )
            ) > 0
        )) AS is_in_recovery
        """
    is_in_recovery = conn.queryscalar(query)
    if not is_in_recovery:
        raise Exception("Instance not in recovery or recovery file is "
                        "missing.")

    if conn.server_version < 120000:
        # Fetch primary_conninfo from recovery.conf file
        # Note: won't work with PG12
        query = """\
        SELECT l FROM unnest(
          string_to_array(
            pg_read_file('recovery.conf'), E'\\n'
          )
        ) AS l
        WHERE l LIKE '%primary\\_conninfo%'
        """
    else:
        query = """\
        SELECT 'primary_conninfo='||quote_literal(setting) AS l
            FROM pg_settings WHERE name='primary_conninfo'
        """
    r = list(conn.query(query))
    if len(r) == 0:
        raise Exception("Unable to get primary_conninfo.")
    pci = r[0]['l']
    # Parse and return primary_conninfo
    return parse_primary_conninfo(pci)


class Probe:
    """Base class for all plugins."""
    # At which level the information is gathered: host, instance or db
    level = None
    # Optionnal name of the probe
    name = None
    # Previous measures used for compute delta
    last_measure = {}
    last_measure_time = {}
    home = None

    def __init__(self, options):
        pass

    def check(self):
        """Check if the plugin can run on the target."""
        pass

    def run(self):
        """Returns the result."""
        pass

    def set_home(self, home):
        self.home = home

    def get_name(self):
        """Computes the name of the probe."""
        # Let the plugin overwrite the name
        if self.name is not None:
            return self.name

        # Compute the name from the class of the plugin
        m = re.search(r'^probe_(\w+)$', self.__class__.__name__.lower())
        if m is not None:
            return m.group(1)

        logger.error("Could not get the name of the probe")
        return None

    def get_last_measure(self, key):
        row = db.get_last_measure(
            self.home,
            'monitoring.db',
            key
        )
        if row:
            return dict(time=row[0], data=json.loads(row[1]))

    def upsert_last_measure(self, time, key, data):
        db.upsert_last_measure(
            self.home,
            'monitoring.db',
            time,
            key,
            data
        )

    def delta(self, key, current_values):
        """
        Compute a delta between measures of two runs.

        Args:
            key (str): identify the values
            current_values (dict): mapping of latest measures

        Returns:
            a tuple of the time interval of the delta in seconds and a
            dict a delta with the same keys as the input.
        """
        current_time = time.time()
        store_key = self.get_name() + key
        last_measure = self.get_last_measure(store_key)
        delta = dict(current=current_values)
        delta_value = None
        # Compute deltas and update last_* variables
        try:
            if last_measure:
                delta['measure_interval'] = current_time - last_measure['time']

                for k in current_values.keys():
                    delta_value = current_values[k] - \
                        last_measure['data'][k]
                    if delta_value < 0:
                        raise Exception('Negative delta value.')
                    delta[k] = delta_value
        except Exception as e:
            logger.debug("Failed to compute delta metric: %s", e)

        # Update/insert last measure for next delta calculation
        self.upsert_last_measure(
            current_time,
            store_key,
            current_values
        )

        return delta

    def __repr__(self):
        return self.get_name()


class HostProbe(Probe):
    """
        system probe base class
    """
    level = 'host'
    system = None  # kernel name from os.uname()[0]
    min_version = None
    max_version = None

    def check(self):
        """Check if the probe can run on this system."""
        if self.system is not None:
            if self.system != os.uname()[0]:
                return False

        version = [int(x) for x
                   in re.sub(r'[-_+].*$', '', os.uname()[2]).split('.')]
        if self.min_version is not None:
            if version[0:len(self.min_version)] < self.min_version:
                return False

        if self.max_version is not None:
            if version[0:len(self.max_version)] > self.max_version:
                return False

        return True


class SqlProbe(Probe):
    """
        postgres probe base class
    """
    min_version = None
    max_version = None
    sql = None
    no_standby = False

    # Delta columns is a list of columns in the output of the query on
    # which delta are computed between to runs. When is it set, a new
    # key is added with the time interval of the delta named with
    # measure_interval. delta_key a column used as unique key to
    # compute delta on multiline output.
    delta_columns = None
    delta_key = None
    timeout = False

    def check(self, version=None):
        """Check if the plugin can run on the target version of PostgreSQL."""
        if version is None:
            return False

        if self.min_version is not None:
            if version < self.min_version:
                return False

        if self.max_version is not None:
            if version > self.max_version:
                return False

        return True

    def run_sql(self, conn, conninfo, sql):
        """Get the result of the SQL query"""
        if sql is None:
            return []

        database = conninfo['dbname']

        output = []
        try:
            if self.timeout:
                conn.execute(
                    "SET statement_timeout = '%ss';", (self.timeout,))

            cluster_name = conninfo['instance'].replace('/', '')
            sql = "-- probe %s\n%s" % (self, sql)
            for r in conn.query(sql):
                # Compute delta if the probe needs that
                if self.delta_columns is not None:
                    to_delta = {}

                    for k in self.delta_columns:
                        if k in r.keys():
                            to_delta[k] = r.pop(k)

                    # Create the store key for the delta
                    if self.delta_key is not None:
                        key = cluster_name + database + r[self.delta_key]
                    else:
                        key = cluster_name + database

                    # Calculate and add delta.
                    r.update(self.delta(key, to_delta))

                output.append(r)
        except Exception as e:
            logger.error(
                "Unable to run probe \"%s\" on \"%s\" on database \"%s\": %s",
                self.get_name(), conninfo['instance'], database, e,
                exc_info=True,
            )
        return output

    def run(self, conn, conninfo):
        """Execute the query depending on the level configured."""
        if self.no_standby and conninfo['standby']:
            return []

        return self.run_sql(conn, conninfo, self.sql)


class probe_sessions(SqlProbe):
    level = 'instance'

    def check(self, version):
        if not super().check(version):
            return False

        if version < 90200:
            self.sql = """select
  d.datname as dbname,
  coalesce(sum((current_query not in ('<IDLE>','<IDLE> in transaction') and not waiting)::integer), 0) as active,
  coalesce(sum(waiting::integer), 0) as waiting,
  coalesce(sum((current_query='<IDLE>')::integer), 0) as idle,
  coalesce(sum((current_query='<IDLE> in transaction')::integer), 0) as idle_in_xact,
  0 as idle_in_xact_aborted,
  coalesce(sum((current_query='<FASTPATH> function call')::integer), 0) as fastpath,
  coalesce(sum((current_query='<command string not enabled>')::integer), 0) as disabled,
  coalesce(sum((current_query='<insufficient privilege>')::integer), 0) as no_priv
from pg_database d
  left join pg_stat_activity a on (d.oid = a.datid)
where d.datallowconn
group by d.datname"""  # noqa
        elif version >= 90200 and version < 90600:
            self.sql = """select
  d.datname as dbname,
  coalesce(sum((state = 'active' and not waiting)::integer), 0) as active,
  coalesce(sum((state = 'active' and waiting)::integer), 0) as waiting,
  coalesce(sum((state = 'idle')::integer), 0) as idle,
  coalesce(sum((state = 'idle in transaction')::integer), 0) as idle_in_xact,
  coalesce(sum((state = 'idle in transaction (aborted)')::integer), 0) as idle_in_xact_aborted,
  coalesce(sum((state = 'fastpath function call')::integer), 0) as fastpath,
  coalesce(sum((state = 'disabled')::integer), 0) as disabled,
  coalesce(sum((query = '<insufficient privilege>')::integer), 0) as no_priv
from pg_database d
  left join pg_stat_activity a on (d.oid = a.datid)
where d.datallowconn
group by d.datname"""  # noqa
        elif version >= 90600:
            self.sql = """select
  d.datname as dbname,
  coalesce(sum((state = 'active' and wait_event_type IS DISTINCT FROM 'Lock')::integer), 0) as active,
  coalesce(sum((state = 'active' and wait_event_type IS NOT DISTINCT FROM 'Lock')::integer), 0) as waiting,
  coalesce(sum((state = 'idle')::integer), 0) as idle,
  coalesce(sum((state = 'idle in transaction')::integer), 0) as idle_in_xact,
  coalesce(sum((state = 'idle in transaction (aborted)')::integer), 0) as idle_in_xact_aborted,
  coalesce(sum((state = 'fastpath function call')::integer), 0) as fastpath,
  coalesce(sum((state = 'disabled')::integer), 0) as disabled,
  coalesce(sum((query = '<insufficient privilege>')::integer), 0) as no_priv
from pg_database d
  left join pg_stat_activity a on (d.oid = a.datid)
where d.datallowconn
group by d.datname"""  # noqa

        return True


class probe_xacts(SqlProbe):
    level = 'instance'
    min_version = 70400
    sql = """select
  s.datname as dbname,
  xact_commit as n_commit,
  xact_rollback as n_rollback
from pg_stat_database s
  join pg_database d on (d.oid = s.datid)
where d.datallowconn"""
    delta_columns = ['n_commit', 'n_rollback']
    delta_key = 'dbname'


class probe_locks(SqlProbe):
    level = 'instance'
    sql = """select
  d.datname as dbname,
  coalesce(sum((mode = 'AccessShareLock' and granted)::integer), 0) as access_share,
  coalesce(sum((mode = 'AccessShareLock' and not granted)::integer), 0) as waiting_access_share,
  coalesce(sum((mode = 'RowShareLock' and granted)::integer), 0) as row_share,
  coalesce(sum((mode = 'RowShareLock' and not granted)::integer), 0) as waiting_row_share,
  coalesce(sum((mode = 'RowExclusiveLock' and granted)::integer), 0) as row_exclusive,
  coalesce(sum((mode = 'RowExclusiveLock' and not granted)::integer), 0) as waiting_row_exclusive,
  coalesce(sum((mode = 'ShareUpdateExclusiveLock' and granted)::integer), 0) as share_update_exclusive,
  coalesce(sum((mode = 'ShareUpdateExclusiveLock' and not granted)::integer), 0) as waiting_share_update_exclusive,
  coalesce(sum((mode = 'ShareLock' and granted)::integer), 0) as share,
  coalesce(sum((mode = 'ShareLock' and not granted)::integer), 0) as waiting_share,
  coalesce(sum((mode = 'ShareRowExclusiveLock' and granted)::integer), 0) as share_row_exclusive,
  coalesce(sum((mode = 'ShareRowExclusiveLock' and not granted)::integer), 0) as waiting_share_row_exclusive,
  coalesce(sum((mode = 'ExclusiveLock' and granted)::integer), 0) as exclusive,
  coalesce(sum((mode = 'ExclusiveLock' and not granted)::integer), 0) as waiting_exclusive,
  coalesce(sum((mode = 'AccessExclusiveLock' and granted)::integer), 0) as access_exclusive,
  coalesce(sum((mode = 'AccessExclusiveLock' and not granted)::integer), 0) as waiting_access_exclusive,
  coalesce(sum((mode = 'SIReadLock' and granted)::integer), 0) as siread
from pg_database d
  left join pg_locks l on (l.database = d.oid)
where d.datallowconn
group by d.datname"""  # noqa


class probe_blocks(SqlProbe):
    level = 'instance'
    sql = """select
  s.datname as dbname,
  blks_read,
  blks_hit,
  coalesce(blks_hit::float*100/nullif(blks_read+blks_hit, 0), 0) as hitmiss_ratio
from pg_stat_database s
  join pg_database d on (d.oid = s.datid)
where d.datallowconn"""  # noqa
    delta_columns = ['blks_read', 'blks_hit']
    delta_key = 'dbname'


class probe_bgwriter(SqlProbe):
    level = 'instance'
    min_version = 80300
    sql = """select * from pg_stat_bgwriter"""
    delta_columns = [
        'checkpoints_timed', 'checkpoints_req', 'checkpoint_write_time',
        'checkpoint_sync_time', 'buffers_checkpoint', 'buffers_clean',
        'maxwritten_clean', 'buffers_backend', 'buffers_backend_fsync',
        'buffers_alloc'
    ]


class probe_db_size(SqlProbe):
    level = 'instance'
    sql = """select
  datname as dbname,
  pg_database_size(oid) as size
from pg_database
where datallowconn"""


class probe_tblspc_size(SqlProbe):
    level = 'instance'
    sql = """select
  spcname,
  pg_tablespace_size(oid) as size
from pg_tablespace"""


class probe_filesystems_size(HostProbe):
    system = 'Linux'

    def run(self):
        return SysInfo().file_systems()


class probe_cpu(HostProbe):
    system = 'Linux'
    hz = os.sysconf(os.sysconf_names['SC_CLK_TCK'])

    def run(self):
        to_delta = {}
        stat = open('/proc/stat')
        for line in stat:
            cols = line.split()
            if len(cols) == 0:
                continue
            if cols[0] == 'cpu':
                # Convert values to int then in milliseconds,
                to_delta = {
                    'time_user': ((int(cols[1]) + int(cols[2])) * 1000 /
                                  self.hz),
                    'time_system': (int(cols[3]) + int(cols[6]) +
                                    int(cols[7])) * 1000 / self.hz,
                    'time_idle': int(cols[4]) * 1000 / self.hz,
                    'time_iowait': int(cols[5]) * 1000 / self.hz,
                    'time_steal': int(cols[8]) * 1000 / self.hz
                }
        stat.close()

        # Compute deltas for values of /proc/stat since boot time
        metrics = self.delta('global', to_delta)
        metrics['cpu'] = 'global'

        return [metrics]


class probe_process(HostProbe):
    system = 'Linux'

    def run(self):
        metrics = {}
        # Process information is partly stored in /proc/stat, ctxt and
        # processes are ever incresing counters, compute deltas on
        # them.
        to_delta = {}
        stat = open('/proc/stat')
        for line in stat:
            cols = line.split()
            if len(cols) == 0:
                continue
            if cols[0] == 'ctxt':
                to_delta['context_switches'] = int(cols[1])
            if cols[0] == 'processes':
                to_delta['forks'] = int(cols[1])
            if cols[0] == 'procs_running':
                metrics['procs_running'] = int(cols[1])
            if cols[0] == 'procs_blocked':
                metrics['procs_blocked'] = int(cols[1])
        stat.close()

        # Total number of process is stored in /proc/loadavg
        load = open('/proc/loadavg')
        m = re.match(r'.*/(\d+) ', load.readline())
        load.close()
        if m:
            metrics['procs_total'] = m.group(1)
        else:
            metrics['procs_total'] = 0

        # Compute deltas for values of /proc/stat since boot time
        metrics.update(self.delta('key', to_delta))
        return [metrics]


class probe_memory(HostProbe):
    system = 'Linux'

    def run(self):
        sysinfo = SysInfo()
        meminfo = sysinfo.mem_info()

        return [{
            'mem_total': meminfo['MemTotal'],
            'mem_used': meminfo['MemTotal'] - meminfo['MemFree'],
            'mem_free': meminfo['MemFree'],
            'mem_buffers': meminfo['Buffers'],
            'mem_cached': meminfo['Cached'],
            'swap_total': meminfo['SwapTotal'],
            'swap_used': meminfo['SwapTotal'] - meminfo['SwapFree']
        }]


class probe_loadavg(HostProbe):
    system = 'Linux'

    def run(self):
        loadavg = open('/proc/loadavg')
        cols = loadavg.readline().split()
        loadavg.close()
        return [{
            'load1': cols[0],
            'load5': cols[1],
            'load15': cols[2]
        }]


class probe_wal_files(SqlProbe):
    level = 'instance'
    min_version = 80200

    def run(self, conn, conninfo):
        version = conn.server_version

        if conninfo['standby']:
            return []

        metric = {}
        if version < 100000:
            sql = """
            SELECT count(s.f) AS total,
                   sum((pg_stat_file('pg_xlog/'||s.f)).size) AS total_size,
                   pg_current_xlog_location() as current_location
            FROM pg_ls_dir('pg_xlog') AS s(f)
            WHERE f ~ E'^[0-9A-F]{24}$'
            """
        else:
            sql = """
            SELECT count(s.f) AS total,
                   sum((pg_stat_file('pg_wal/'||s.f)).size) AS total_size,
                   pg_current_wal_lsn() as current_location
            FROM pg_ls_dir('pg_wal') AS s(f)
            WHERE f ~ E'^[0-9A-F]{24}$'
            """
        rows = self.run_sql(conn, conninfo, sql)

        metric['total'] = rows[0]['total']
        metric['total_size'] = rows[0]['total_size']
        metric['current_location'] = rows[0]['current_location']

        if version < 100000:
            sql = r"""
            SELECT count(s.f) AS archive_ready
            FROM pg_ls_dir('pg_xlog/archive_status') AS s(f)
            WHERE f ~ E'\.ready$'
            """  # noqa W605
        else:
            sql = r"""
            SELECT count(s.f) AS archive_ready
            FROM pg_ls_dir('pg_wal/archive_status') AS s(f)
            WHERE f ~ E'\.ready$'
            """  # noqa W605
        rows = self.run_sql(conn, conninfo, sql)

        metric['archive_ready'] = rows[0]['archive_ready']

        # Calcutate the written size by using the delta between the
        # position between to runs. The current xlog location must be
        # converted to an number first
        m = re.match(r'^([0-9A-F]+)/([0-9A-F]+)$', metric['current_location'])
        if m:
            current = int("0xff000000", 0) * \
                int("0x" + m.group(1), 0) + int("0x" + m.group(2), 0)
        else:
            logger.error("Unable to convert xlog location to a number")
            return []

        metric.update(self.delta(
            conninfo['instance'].replace('/', ''),
            {'written_size': current},
        ))
        return [metric]


class probe_replication_lag(SqlProbe):
    # Streaming replication lag probe, in bytes
    #
    # The instance is in recovery mode ?
    # + recovery.conf file exists ?
    # + recovery.conf contains primary_conninfo ?
    # + get primary node's current wal position with IDENTIFY_SYSTEM
    #   through streaming repl. protocol
    # + return primary wal position and current replay diff.
    level = 'instance'
    min_version = 90000

    def run(self, conn, conninfo):
        if not conninfo['standby']:
            return []

        try:
            # Get primary parameters from primary_conninfo
            dsn = get_primary_conninfo(conn)

            # Let's fetch primary current wal position with IDENTIFY_SYSTEM
            # through streaming replication protocol.
            with closing(psycopg2.connect(
                dsn, connection_factory=PhysicalReplicationConnection
            )) as p_conn:
                with p_conn.cursor() as p_cur:
                    p_cur.execute("IDENTIFY_SYSTEM")
                    rows = p_cur.fetchall()

                    if len(rows) == 0:
                        return []
                    xlogpos = rows[0][2]

            # Proceed with LSN diff
            if conn.server_version >= 100000:
                # Use pg_wal_lsn_diff(pg_lsn, pg_lsn)
                rows = conn.query("""\
                SELECT pg_wal_lsn_diff(
                    '{xlogpos}'::pg_lsn,
                    pg_last_wal_replay_lsn()
                ) AS lsn_diff
                """.format(xlogpos=xlogpos))
            elif conn.server_version >= 90500:
                # Use pg_xlog_lsn_diff(pg_lsn, pg_lsn)
                rows = conn.query("""\
                SELECT pg_xlog_location_diff(
                    '{xlogpos}'::pg_lsn,
                    pg_last_xlog_replay_location()
                ) AS lsn_diff
                """.format(xlogpos=xlogpos))
            else:
                # Use pg_xlog_lsn_diff(TEXT, TEXT)
                rows = conn.query("""\
                SELECT pg_xlog_location_diff(
                    '{xlogpos}'::TEXT,
                    pg_last_xlog_replay_location()::TEXT
                ) AS lsn_diff
                """.format(xlogpos=xlogpos))
            r = list(rows)
            if len(r) == 0:
                return []
            return [{'lag': int(r[0]['lsn_diff'])}]

        except Exception as e:
            logger.exception(str(e))
            return []


class probe_temp_files_size_delta(SqlProbe):
    # Temporary files probe
    level = 'instance'
    sql = """SELECT
  s.datname AS dbname,
  temp_bytes AS size
FROM pg_stat_database s
JOIN pg_database d ON (d.oid = s.datid)
WHERE d.datallowconn"""  # noqa
    delta_columns = ['size']
    delta_key = 'dbname'


class probe_replication_connection(SqlProbe):
    # Check if the instance is connected to streaming replication according to
    # primary_conninfo from recovery.conf file.
    level = 'instance'
    min_version = 96000

    def run(self, conn, conninfo):
        if not conninfo['standby']:
            return []

        try:
            # Get primary parameters from primary_conninfo
            dsn = parse_dsn(get_primary_conninfo(conn))
            p_host = dsn['host']

            # pg_stat_wal_receiver lookup
            rows = conn.query("""\
            SELECT '{p_host}' AS upstream,
            CASE WHEN COUNT(*) > 0 THEN 1 ELSE 0 END AS connected
            FROM pg_stat_wal_receiver
            WHERE status='streaming' AND
                    conninfo LIKE '%host={p_host}%'
            """.format(p_host=p_host))
            return list(rows)
        except Exception as e:
            logger.exception(str(e))
            return []


class probe_heap_bloat(SqlProbe):
    # Heap bloat estimation probe
    # Query coming from https://github.com/ioguix/pgsql-bloat-estimation/
    level = 'database'
    sql = """
SELECT current_database() AS dbname,
  SUM(bloat_size)::FLOAT/SUM(bs*tblpages)::FLOAT*100 AS ratio
FROM (
  SELECT
  CASE WHEN tblpages - est_tblpages_ff > 0 THEN (tblpages-est_tblpages_ff)*bs ELSE 0 END AS bloat_size,
  bs, tblpages
  FROM (
    SELECT
      ceil( reltuples / ( (bs-page_hdr)*fillfactor/(tpl_size*100) ) ) + ceil( toasttuples / 4 ) AS est_tblpages_ff,
      tblpages, bs
    FROM (
      SELECT
        ( 4 + tpl_hdr_size + tpl_data_size + (2*ma)
          - CASE WHEN tpl_hdr_size%ma = 0 THEN ma ELSE tpl_hdr_size%ma END
          - CASE WHEN ceil(tpl_data_size)::int%ma = 0 THEN ma ELSE ceil(tpl_data_size)::int%ma END
        ) AS tpl_size, bs - page_hdr AS size_per_block, (heappages + toastpages) AS tblpages, heappages,
        toastpages, reltuples, toasttuples, bs, page_hdr, tblid, schemaname, tblname, fillfactor
      FROM (
        SELECT
          tbl.oid AS tblid, ns.nspname AS schemaname, tbl.relname AS tblname,
          tbl.reltuples, tbl.relpages AS heappages, coalesce(toast.relpages, 0) AS toastpages,
          coalesce(toast.reltuples, 0) AS toasttuples,
          coalesce(substring(
            array_to_string(tbl.reloptions, ' ')
            FROM 'fillfactor=([0-9]+)')::smallint, 100) AS fillfactor,
          current_setting('block_size')::numeric AS bs,
          CASE WHEN version()~'mingw32' OR version()~'64-bit|x86_64|ppc64|ia64|amd64' THEN 8 ELSE 4 END AS ma,
          24 AS page_hdr,
          23 + CASE WHEN MAX(coalesce(s.null_frac,0)) > 0 THEN ( 7 + count(s.attname) ) / 8 ELSE 0::int END
           + CASE WHEN bool_or(att.attname = 'oid' and att.attnum < 0) THEN 4 ELSE 0 END AS tpl_hdr_size,
          sum( (1-coalesce(s.null_frac, 0)) * coalesce(s.avg_width, 0) ) AS tpl_data_size
        FROM pg_attribute AS att
          JOIN pg_class AS tbl ON att.attrelid = tbl.oid
          JOIN pg_namespace AS ns ON ns.oid = tbl.relnamespace
          LEFT JOIN pg_stats AS s ON s.schemaname=ns.nspname
            AND s.tablename = tbl.relname AND s.inherited=false AND s.attname=att.attname
          LEFT JOIN pg_class AS toast ON tbl.reltoastrelid = toast.oid
        WHERE NOT att.attisdropped
          AND tbl.relkind = 'r'
        GROUP BY 1,2,3,4,5,6,7,8,9,10
        ORDER BY 2,3
      ) AS s
    ) AS s2
  ) AS s3
) AS s4;
    """  # noqa


class probe_btree_bloat(SqlProbe):
    # Btree index bloat estimation probe
    timeout = 30
    level = 'database'
    sql = """
SELECT
  current_database() AS dbname,
  SUM(bloat_size) / SUM(indexes_size) * 100 AS ratio
FROM (
  SELECT
    SUM(pg_relation_size(quote_ident(schemaname) || '.' || quote_ident(indexname)))::BIGINT AS indexes_size
  FROM pg_catalog.pg_indexes
  GROUP BY schemaname
) AS indexes,
(
  SELECT
    SUM(bloat_size) AS bloat_size,
    schemaname
  FROM (
    %s
  ) AS a
  GROUP BY schemaname
) AS ibloat
WHERE schemaname !~ '^pg_temp'
AND schemaname !~ '^pg_toast';
"""  % (INDEX_BTREE_BLOAT_SQL) # noqa



# User Input Validator
#
# These functions provide stateless validation of user input, mainly CLI
# arguments and environment variables.
#
# On invalid input, a ValueError is raised. Other exceptions are considered a
# bug.
#
# A validator is idempotent. It must accepts what it returns.

import json
import logging
import os.path
import re
from logging.handlers import SysLogHandler

from .log import HANDLERS as LOG_METHODS
from .pycompat import urlparse
from .utils import strtobool


_address_re = re.compile(
    r'(?:[3-9]\d?|2(?:5[0-5]|[0-4]?\d)?|1\d{0,2}|\d)'
    r'(\.(?:[3-9]\d?|2(?:5[0-5]|[0-4]?\d)?|1\d{0,2}|\d'
    r')){3}$'
)


def address(raw):
    if not _address_re.match(raw):
        raise ValueError('invalid address')
    return raw


def boolean(raw):
    if raw in (True, False):
        return raw

    return strtobool(raw)


def dir_(raw):
    raw = os.path.realpath(raw)
    if not os.path.isdir(raw):
        raise ValueError('Not a directory')
    return raw


def file_(raw):
    if not raw:
        return raw
    raw = os.path.realpath(raw)
    if not os.path.exists(raw):
        raise ValueError('%s: File not found' % raw)
    return raw


def path(raw):
    if not raw:
        return raw
    raw = os.path.realpath(raw)
    parent = os.path.dirname(raw)
    if not os.path.isdir(parent):
        raise ValueError("Missing parent directory of: %s" % raw)
    return raw


def fqdn(raw):
    if '\n' in raw:
        raise ValueError("New line in FQDN.")
    if not re.match(
        r"(?=^.{1,253}$)"  # check length between 4 and 253
        r"(^((?!-)[a-zA-Z0-9-]{1,63}(?<!-))"
        r"(\.(?!-)[a-zA-Z0-9-]{1,63}(?<!-))*$)",
        raw,
    ):
        raise ValueError("%s is not an FQDN" % raw)
    return raw


_identifier_re = re.compile(r'^[a-zA-Z0-9]+$')


def jsonlist(raw):
    if hasattr(raw, 'lower'):
        raw = json.loads(raw)

    if not isinstance(raw, list):
        raise ValueError('not a list')

    raw = [str(e) for e in raw]
    for entry in raw:
        if not _identifier_re.match(entry):
            raise ValueError('%s is invalid' % entry)

    return raw


def port(raw):
    port = int(raw)

    if 0 > port or port > 65635:
        raise ValueError('Port out of range')

    return port


def loglevel(raw):
    raw = raw.upper()
    levelnames = list()
    if hasattr(logging, '_levelNames'):  # pragma: nocover_py3
        levelnames = logging._levelNames
    elif hasattr(logging, '_nameToLevel'):  # pragma: nocover_py2
        levelnames = logging._nameToLevel
    if raw not in levelnames:
        raise ValueError('unkown log level')
    return raw


def logmethod(raw):
    if raw not in LOG_METHODS:
        raise ValueError('unknown logging method %s' % (raw,))
    return raw


def syslogfacility(raw):
    if raw not in SysLogHandler.facility_names:
        raise ValueError('unkown syslog facility')
    return raw


def writeabledir(raw):
    raw = dir_(raw)
    if not os.access(raw, os.W_OK):
        raise ValueError('Not writable')
    return raw


def commalist(raw):
    return list(filter(None, [w.strip() for w in raw.split(',')]))


def nday(raw):
    nday = int(raw)

    if nday < 1:
        raise ValueError('Number of day not valid')

    return nday


def url(raw):
    url = urlparse(raw)
    if not url.scheme.startswith('http'):
        raise ValueError("HTTP URL required")
    if not url.netloc:
        raise ValueError("Missing host and port")
    return raw



import logging
import time

from bottle import Bottle, default_app, request

from temboardagent.tools import validate_parameters
from temboardagent.command import (
    oneline_cmd_to_array,
    exec_script,
)
from temboardagent.toolkit.configuration import OptionSpec
from temboardagent.notification import NotificationMgmt, Notification

from .types import T_CONTROL


bottle = Bottle()
logger = logging.getLogger(__name__)


@bottle.post('/control')
def post_pg_control(pgconn):
    """Start, stop and restart Postgres instance."""
    app = default_app().temboard
    # Control instance
    validate_parameters(request.json, [
        ('action', T_CONTROL, False)
    ])
    action = request.json['action']
    logger.info("PostgreSQL '%s' requested." % action)
    NotificationMgmt.push(app.config,
                          Notification(username=request.username,
                                       message="PostgreSQL %s" % action))

    cmd = app.config.administration.pg_ctl % action
    cmd_args = oneline_cmd_to_array(cmd)
    (rcode, stdout, stderr) = exec_script(cmd_args)
    if rcode != 0:
        raise Exception(str(stderr))
    # Let's check if PostgreSQL is up & running after having executed
    # 'start' or 'restart' action.
    if action in ['start', 'restart']:
        # When a start/restart operation is requested, after the
        # startup/pg_ctl script has been executed then we check that
        # postgres is up & running:
        # while the PG conn. is not working then, for 10 seconds (max)
        # we'll check (connect/SELECT 1/disconnect) the connection, every
        # 0.5 second.
        retry = True
        t_start = time.time()
        while retry:
            try:
                pgconn.execute('SELECT 1')
                logger.info("Done.")
                return dict(action=action, state='ok')
            except Exception:
                if (time.time() - t_start) > 10:
                    logger.info("Failed.")
                    return dict(action=action, state='ko')
            logger.info("Retrying...")
            time.sleep(0.5)

    elif action == 'stop':
        # Check the PG conn is not working anymore.
        try:
            retry = True
            t_start = time.time()
            while retry:
                pgconn.execute('SELECT 1')
                time.sleep(0.5)
                if (time.time() - t_start) > 10:
                    retry = False
            logger.info("Failed.")
            return dict(action=action, state='ko')
        except Exception:
            logger.info("Done.")
            return dict(action=action, state='ok')

    elif action == 'reload':
        logger.info("Done.")
        return dict(action=action, state='ok')


class AdministrationPlugin:
    PG_MIN_VERSION = (90400, 9.4)
    options_specs = [
        OptionSpec('administration', 'pg_ctl', default=None),
    ]

    def __init__(self, app, **kw):
        self.app = app
        self.app.config.add_specs(self.options_specs)

    def load(self):
        default_app().mount('/administration/', bottle)

    def unload(self):
        self.app.config.remove_specs(self.options_specs)



# coding: utf-8

from datetime import datetime
import json
import logging
import sqlite3
from textwrap import dedent

from ..taskmanager import Task
from ..errors import StorageEngineError

logger = logging.getLogger(__name__)


def epoch_to_datetime(epoch):
    if epoch:
        return datetime.utcfromtimestamp(epoch)


def datetime_to_epoch(timestamp):
    if timestamp:
        return int((timestamp - datetime(1970, 1, 1)).total_seconds())
    return 0


class TaskListSQLite3Engine(object):
    """SQLite3 storage engine for task list management."""

    def __init__(self, dbpath):
        self.dbpath = dbpath
        self.conn = sqlite3.connect(self.dbpath)

    def bootstrap(self):
        try:
            with self.conn:
                c = self.conn.cursor()
                c.execute("PRAGMA synchronous = 1")
                c.execute(
                    dedent("""
                        CREATE TABLE IF NOT EXISTS tasks (
                            id TEXT PRIMARY KEY,
                            worker_name TEXT,
                            start_datetime BIGINT,
                            stop_datetime BIGINT,
                            status SMALLINT,
                            output TEXT,
                            options TEXT,
                            redo_interval INTEGER DEFAULT 0 NOT NULL,
                            expire INTEGER DEFAULT 0 NOT NULL
                        )
                    """)
                )
        except sqlite3.OperationalError as e:
            logger.exception(str(e))
            raise StorageEngineError("Could not bootstrap storage engine.")

    def insert(self, task):
        try:
            with self.conn:
                c = self.conn.cursor()
                c.execute(
                    "INSERT INTO tasks VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?)",
                    (
                        task.id,
                        task.worker_name,
                        datetime_to_epoch(task.start_datetime),
                        datetime_to_epoch(task.stop_datetime),
                        task.status,
                        str(task.output) if task.output else None,
                        json.dumps(task.options),
                        task.redo_interval or 0,
                        task.expire or 0
                    )
                )
        except sqlite3.IntegrityError:
            raise KeyError("Task with id=%s already exists" % task.id)
        except sqlite3.Error as e:
            logger.exception(str(e))
            raise StorageEngineError("Could not insert task.")

    def update(self, task):
        try:
            with self.conn:
                c = self.conn.cursor()
                c.execute(
                    dedent("""
                        UPDATE tasks
                        SET
                          worker_name = ?,
                          start_datetime = ?,
                          stop_datetime = ?,
                          status = ?,
                          output = ?,
                          options = ?,
                          redo_interval = ?,
                          expire = ?
                        WHERE id = ?
                    """),
                    (
                        task.worker_name,
                        datetime_to_epoch(task.start_datetime),
                        datetime_to_epoch(task.stop_datetime),
                        task.status,
                        str(task.output) if task.output else None,
                        json.dumps(task.options),
                        task.redo_interval or 0,
                        task.expire or 0,
                        task.id
                    )
                )
        except sqlite3.Error as e:
            logger.exception(str(e))
            raise StorageEngineError("Could not update task.")

    def delete(self, id):
        try:
            with self.conn:
                c = self.conn.cursor()
                c.execute(
                    "DELETE FROM tasks WHERE id = ?",
                    (id,)
                )
        except sqlite3.Error as e:
            logger.exception(str(e))
            raise StorageEngineError("Could not delete task with id=%s" % id)

    def get(self, id):
        try:
            with self.conn:
                c = self.conn.cursor()
                c.execute(
                    dedent("""
                        SELECT
                            id, worker_name, start_datetime, stop_datetime,
                            status, output, options, redo_interval, expire
                        FROM tasks
                        WHERE id = ?
                    """),
                    (id,)
                )
                r = c.fetchone()
                if not r:
                    return

                try:
                    options = json.loads(r[6])
                except ValueError as e:
                    logger.exception(str(e))
                    logger.debug(r)
                    raise StorageEngineError("Could not decode task options.")

                return Task(
                    id=r[0],
                    worker_name=r[1],
                    start_datetime=epoch_to_datetime(r[2]),
                    stop_datetime=epoch_to_datetime(r[3]),
                    status=r[4],
                    output=r[5],
                    options=options,
                    redo_interval=r[7],
                    expire=r[8]
                )

        except sqlite3.Error as e:
            logger.exception(str(e))
            raise StorageEngineError("Could not get task with id=%s" % id)

    def list(self):
        try:
            with self.conn:
                c = self.conn.cursor()
                c.execute(
                    dedent("""
                        SELECT
                            id, worker_name, start_datetime, stop_datetime,
                            status, output, options, redo_interval, expire
                        FROM tasks
                        ORDER BY start_datetime ASC
                    """)
                )
                for r in c.fetchall():

                    try:
                        options = json.loads(r[6])
                    except ValueError as e:
                        logger.exception(str(e))
                        logger.error("Could not convert row to Task.")
                        logger.debug(r)
                        continue

                    yield Task(
                        id=r[0],
                        worker_name=r[1],
                        start_datetime=epoch_to_datetime(r[2]),
                        stop_datetime=epoch_to_datetime(r[3]),
                        status=r[4],
                        output=r[5],
                        options=options,
                        redo_interval=r[7],
                        expire=r[8]
                    )

        except sqlite3.Error as e:
            logger.exception(str(e))
            raise StorageEngineError("Could not get task list.")

    def exists(self, id):
        try:
            with self.conn:
                c = self.conn.cursor()
                c.execute("SELECT 1 FROM tasks WHERE id = ?", (id,))
                return (c.fetchone() is not None)
        except sqlite3.Error as e:
            logger.exception(str(e))
            raise StorageEngineError(
                "Could not check that task with id=%s exists" % id
            )

    def count_by_status(self, status):
        try:
            with self.conn:
                c = self.conn.cursor()
                c.execute(
                    "SELECT COUNT(id) FROM tasks WHERE status & ?",
                    (status,)
                )
                return c.fetchone()[0]
        except sqlite3.Error as e:
            logger.exception(str(e))
            raise StorageEngineError(
                "Could not count tasks with status & %s" % status
            )

    def recover(self, st_doing, st_aborted, st_scheduled, st_default, now):
        try:
            with self.conn:
                c = self.conn.cursor()
                # Flag ongoing task as aborted
                c.execute(
                    dedent("""
                        UPDATE tasks SET status = ?, stop_datetime = ?
                        WHERE status & ?
                    """),
                    (st_aborted, datetime_to_epoch(now), st_doing)
                )
                # Reset scheduled task and recurrent tasks to default
                c.execute(dedent("""
                UPDATE tasks SET status = ?
                WHERE status & ? OR redo_interval > 0
                """), (st_default, st_scheduled))
        except sqlite3.Error as e:
            logger.exception(str(e))
            raise StorageEngineError("Could not recover tasks.")

    def list_to_do(self, status, now, redo=False):
        query = dedent("""
                SELECT
                    id, worker_name, start_datetime, stop_datetime,
                    status, output, options, redo_interval, expire
                FROM tasks
                WHERE
        """)
        if redo:
            query += dedent("""
                    redo_interval > 0
                    AND (start_datetime + redo_interval) < ?
                    AND status & ?
            """)
        else:
            query += dedent("""
                    start_datetime < ?
                    AND status & ?
            """)

        try:
            with self.conn:
                c = self.conn.cursor()
                c.execute(query, (datetime_to_epoch(now), status,))
                for r in c.fetchall():

                    try:
                        options = json.loads(r[6])
                    except ValueError as e:
                        logger.exception(str(e))
                        logger.error("Could not convert row to Task.")
                        logger.debug(r)
                        continue

                    yield Task(
                        id=r[0],
                        worker_name=r[1],
                        start_datetime=epoch_to_datetime(r[2]),
                        stop_datetime=epoch_to_datetime(r[3]),
                        status=r[4],
                        output=r[5],
                        options=options,
                        redo_interval=r[7],
                        expire=r[8]
                    )

        except sqlite3.Error as e:
            logger.exception(str(e))
            raise StorageEngineError("Could not get task list.")

    def flush(self):
        try:
            with self.conn:
                c = self.conn.cursor()
                c.execute("DELETE FROM tasks")
                return c.rowcount
        except sqlite3.Error as e:
            logger.exception(str(e))
            raise StorageEngineError("Could not purge task list.")

    def purge(self, status, now):
        try:
            with self.conn:
                c = self.conn.cursor()
                c.execute(
                    dedent("""
                    DELETE FROM tasks
                    WHERE
                        redo_interval = 0
                        AND (stop_datetime + expire) < ?
                        AND status & ?
                    """),
                    (datetime_to_epoch(now), status)
                )
        except sqlite3.Error as e:
            logger.exception(str(e))
            raise StorageEngineError("Could not purge task list.")

    def vacuum(self):
        try:
            with self.conn:
                c = self.conn.cursor()
                c.execute("VACUUM")
        except sqlite3.Error as e:
            logger.exception(str(e))
            raise StorageEngineError("Could not vacuum the database.")



import logging

from bottle import (
    default_app, get, post, request, response, HTTPError, HTTPResponse
)

from ..notification import NotificationMgmt
from ..toolkit.signing import canonicalize_request, verify_v1, InvalidSignature


logger = logging.getLogger(__name__)


@get('/error/<error_code:int>', skip=['signature'])
def get_error(error_code):
    if error_code < 400:
        raise HTTPResponse('HTTPResponse raised', error_code)
    elif error_code < 500:
        raise HTTPError(error_code, 'HTTPError raised')
    else:
        raise Exception("This is an Exception %s" % error_code)


@get('/discover', skip=['signature'])
def get_discover():
    app = default_app().temboard
    discover = app.discover.ensure_latest().copy()
    discover['signature_status'] = 'enabled'

    signature = request.headers.get('x-temboard-signature')
    if signature:
        crequest = canonicalize_request(
            request.method,
            request.path,
            request.headers,
        )
        try:
            if not signature.startswith('v1:'):
                raise InvalidSignature("Unsupported signature version.")
            signature = signature[3:]

            verify_v1(
                app.config.signing_key, signature, crequest)
            discover['signature_status'] = 'valid'
        except InvalidSignature:
            discover['signature_status'] = 'invalid'

    response.set_header('ETag', app.discover.etag)
    return discover


@post('/discover')
def post_discover(pgconn):
    app = default_app().temboard
    data = app.discover.refresh(pgconn).copy()
    # POST endpoint does not bypass signature verification.
    data['signature_status'] = 'valid'
    response.set_header('ETag', app.discover.etag)
    return data


@get
def profile():
    return {'username': request.username, 'signature': 'valid'}


@get('/notifications')
def notifications():
    config = default_app().temboard.config
    return list(NotificationMgmt.get_last_n(config, -1))


@get('/status', skip=['signature'])
def get_status(pgconn):
    app = default_app().temboard
    response.set_header('X-TemBoard-Discover-ETag', app.discover.etag)
    return app.status.get(pgconn)



import logging

from bottle import Bottle, default_app, request

from . import functions as activity_functions
from ...notification import NotificationMgmt, Notification
from ...tools import validate_parameters
from ...queries import QUERIES


bottle = Bottle()
logger = logging.getLogger(__name__)


@bottle.get('/')
def get_activity(pgconn):
    limit = int(request.query.get('limit', 300))
    return activity_functions.get_activity(pgconn, limit)


@bottle.get('/waiting')
def get_activity_waiting(pgconn):
    limit = int(request.query.get('limit', 300))
    return activity_functions.get_activity_waiting(pgconn, limit)


@bottle.get('/blocking')
def get_activity_blocking(pgconn):
    limit = int(request.query.get('limit', 300))
    return activity_functions.get_activity_blocking(pgconn, limit)


@bottle.post('/kill')
def post_activity_kill(pgconn):
    app = default_app().temboard
    validate_parameters(request.json, [
        ('pids', int, True)
    ])
    backends = []
    for pid in request.json['pids']:
        killed = pgconn.queryscalar(
            "SELECT pg_terminate_backend(%s) AS killed" % (pid))
        backends.append(dict(pid=pid, killed=killed))

        # Push a notification.
        try:
            NotificationMgmt.push(
                app.config,
                Notification(
                    username=request.username,
                    message="Backend %s terminated" % (pid)
                )
            )
        except Exception:
            logger.exception("Failed to push notification:")

    return dict(backends=backends)


LOCKS_COLUMNS = [
    'pid',
    'locktype',
    'database',
    'schema',
    'relation',
    'mode',
    'granted',
    'blocking_pid',
    'waitstart',
    'waiting_pids',
]


@bottle.get('/locks')
def get_locks(pgconn):
    # Returns waiting and blocking locks

    discover = default_app().temboard.discover.ensure_latest()
    if discover['postgres']['version_num'] < 100000:
        try:
            LOCKS_COLUMNS.remove('waitstart')
        except ValueError:
            pass

    limit = int(request.query.get('limit', 300))
    query = QUERIES['activity-locks'] + " LIMIT %d" % limit

    return dict(
        columns=SESSION_COLUMNS,
        rows=[row for row in pgconn.query(query)],
    )


SESSION_COLUMNS = [
    'pid',
    'client_addr',
    'username',
    'application_name',
    'database',
    'state',
    'backend_start',
    'query_start',
    'duration',
    'query',
    'waiting',
    'blocking',
]


@bottle.get('/sessions')
def get_sessions(pgconn):
    # List client session with blocking informations.

    limit = int(request.query.get('limit', 300))
    query = QUERIES['activity-sessions'] + " LIMIT %d" % limit
    rows = []
    for row in pgconn.query(query):
        try:
            data = read_proc_info(row['pid'])
            row['cpu_time'] = sum((
                data['utime'],
                data['stime'],
                data['cutime'],
                data['cstime'],
            ))
            row['memory'] = data['VmRSS']
            row['proc_state'] = data['state']
            row['proc_read'] = data['read_bytes']
            row['proc_write'] = data['write_bytes']
        except OSError as e:
            logger.debug("Failed to read /proc/%s info: %s.", row['pid'], e)
            row.update(dict(
                cpu_time=None,
                memory=None,
                proc_read=None,
                proc_state=None,
                proc_write=None,
            ))

        rows.append(row)

    return dict(
        columns=SESSION_COLUMNS,
        rows=rows,
    )


def read_proc_info(pid):
    procdir = '/proc/%d' % pid
    data = dict()
    with open(procdir + '/status') as fo:
        data.update(parse_proc_human(fo, 'VmRSS'))

    with open(procdir + '/stat') as fo:
        data.update(parse_proc_stat(fo.read()))
    with open(procdir + '/io') as fo:
        data.update(parse_proc_human(
            fo,
            'read_bytes', 'write_bytes',
        ))
    return data


def parse_proc_stat(raw):
    values = raw.split()
    return dict(
        state=values[2],
        utime=int(values[13]),
        stime=int(values[14]),
        cutime=int(values[15]),
        cstime=int(values[16]),
    )


def parse_proc_human(fo, *keys):
    # Parse human readable proc format
    data = dict()
    for line in fo:
        key, value = line.split(':')
        if keys and key not in keys:
            continue
        value = value.strip()
        if value.endswith(' kB'):
            value = int(value[:-3]) * 1024
        try:
            value = int(value)
        except ValueError:
            pass
        data[key] = value
    return data


class ActivityPlugin:
    PG_MIN_VERSION = (90400, 9.4)

    def __init__(self, app, **kw):
        self.app = app

    def load(self):
        default_app().mount('/activity', bottle)



import logging
import os

from ..daemon import daemonize
from ..plugins.monitoring import db
from ..toolkit.app import SubCommand
from ..toolkit.services import ServicesManager
from .app import app


logger = logging.getLogger(__name__)


@app.command
class Serve(SubCommand):
    """Combined web and worker services."""

    is_service = True

    def main(self, args):
        # Purge all legacy data queues
        home = self.app.config.temboard.home
        if os.path.exists(home):
            [os.remove(os.path.join(home, f))
             for f in os.listdir(home) if f.endswith('.q')]

        if 'monitoring' in self.app.config.temboard.plugins:
            logger.info("Resetting monitoring data.")
            db.bootstrap(self.app.config.temboard.home, 'monitoring.db')

        self.app.config.load_signing_key()
        self.app.discover.refresh()
        self.app.discover.write()

        if self.app.config.temboard.daemonize:
            daemonize(self.app.config.temboard.pidfile)

        self.app.httpd.services = services = ServicesManager()
        services.add(self.app.worker_pool)
        services.add(self.app.scheduler)

        logger.info("Starting main process.")

        with services:
            self.app.httpd.run()

        return 0



import logging

from .postgres import extract_conninfo_fields
from .queries import QUERIES

logger = logging.getLogger(__name__)


class Status(object):
    def __init__(self, app):
        self.app = app
        self.data = dict(
            postgres=dict(available=True),
            temboard=dict(),
            system=dict(),
        )

    def get(self, conn=None):
        self.data["postgres"] = {}
        self.data["system"] = {}
        self.data["temboard"] = {}
        discover_data = self.app.discover.ensure_latest()
        with self.app.postgres.maybe_connect(conn=conn) as conn:
            logger.debug("Inspecting Postgres instance.")
            collect_postgres(
                self.data,
                conn,
                discover_data,
            )
        self.data["temboard"]["status"] = "running"
        self.data["temboard"]["pid"] = self.app.pid
        self.data["temboard"]["start_time"] = self.app.start_datetime.strftime(
            "%Y-%m-%dT%H:%M:%S%Z"
        )
        self.data["system"]["status"] = "running"
        self.data["system"]["start_time"] = discover_data["system"][
            "start_time"
        ]
        return self.data


def collect_postgres(data, conn, discover_data):
    data["postgres"]["status"] = "running"
    data["postgres"]["pid"] = discover_data["postgres"]["pid"]
    data["postgres"]["start_time"] = discover_data["postgres"]["start_time"]
    try:
        row = conn.queryone(QUERIES["status"])
        data["postgres"].update(row)
        conninfo = extract_conninfo_fields(
            data["postgres"]["primary_conninfo"]
        )
        data["postgres"]["primary"] = conninfo
        data["postgres"]["primary_conninfo"] = " ".join(
            "{}={}".format(k, v) for k, v in conninfo.items()
        )
    except Exception as e:
        logger.warning("Cannot collect extra Postgres info: %s", e)



from .toolkit.errors import UserError

__all__ = ['UserError']


class NotificationError(Exception):
    """ Notification errors """
    def __init__(self, message):
        Exception.__init__(self, message)
        self.message = str(message)



import json
import time
import os
import re

from . import db
from ...notification import NotificationMgmt
from ...inventory import SysInfo


def get_metrics(app, pool=None):
    res = dict()
    pool = pool or app.postgres.pool()
    discover = app.discover.ensure_latest()
    for attempt in pool.auto_reconnect():
        with attempt() as conn:
            dm = DashboardMetrics(conn)
            pgdiscover = discover['postgres']
            res.update(dict(
                buffers=dm.get_buffers(),
                hitratio=dm.get_hitratio(),
                active_backends=dm.get_active_backends(),
                max_connections=pgdiscover['max_connections'],
                databases=dm.get_stat_db(),
                pg_start_time=dm.get_pg_start_time(),
                pg_version=pgdiscover['version'],
                pg_data=pgdiscover['data_directory'],
                pg_port=pgdiscover['port'],
            ))

    sysdiscover = discover['system']
    dm = DashboardMetrics()
    res.update(dict(
        cpu=dm.get_cpu_usage(),
        loadaverage=dm.get_load_average(),
        memory=dm.get_memory_usage(),
        notifications=dm.get_notifications(app.config),
    ))

    sysinfo = SysInfo()

    cpu_models = [cpu['model_name'] for cpu in sysinfo.cpu_info()['cpus']]
    cpu_models_counter = {}
    for elem in cpu_models:
        cpu_models_counter[elem] = cpu_models_counter.get(elem, 0) + 1

    res.update(dict(
        hostname=sysdiscover['fqdn'],
        os_version=sysdiscover['os_version'],
        linux_distribution=sysdiscover['distribution'],
        cpu_models=cpu_models_counter,
        n_cpu=sysdiscover['cpu_count'],
        timestamp=time.time()
    ))
    return res


def get_metrics_queue(config):
    dm = DashboardMetrics()
    msg = dict()

    row = db.get_last_metric(
        config.temboard.home,
        'dashboard.db'
    )
    if row:
        msg = json.loads(row[0])

    msg['notifications'] = dm.get_notifications(config)
    return msg


def get_history_metrics_queue(config):
    return [
        json.loads(d)
        for d, in db.get_all_metrics(
            config.temboard.home,
            'dashboard.db'
        )
    ]


def get_buffers(conn):
    dm = DashboardMetrics(conn)
    return dict(buffers=dm.get_buffers())


def get_hitratio(conn):
    dm = DashboardMetrics(conn)
    return dict(hitratio=dm.get_hitratio())


def get_active_backends(conn):
    dm = DashboardMetrics(conn)
    return dict(active_backends=dm.get_active_backends())


def get_cpu_usage():
    dm = DashboardMetrics()
    return dict(cpu=dm.get_cpu_usage())


def get_loadaverage():
    dm = DashboardMetrics()
    return dict(loadaverage=dm.get_load_average())


def get_memory_usage():
    dm = DashboardMetrics()
    return dict(memory=dm.get_memory_usage())


def get_hostname(config):
    sysinfo = SysInfo()
    return dict(hostname=sysinfo.hostname(config.temboard.hostname))


def get_databases(conn):
    dm = DashboardMetrics(conn)
    return dict(databases=dm.get_stat_db())


class DashboardMetrics:
    conn = None
    config = None
    _instance = None

    def __init__(self, conn=None):
        self.conn = conn

    def get_buffers(self,):
        current_time = time.time()
        current_buffers = self._get_current_buffers()
        return {'nb': current_buffers,
                'time': current_time}

    def get_hitratio(self,):
        return self.conn.queryscalar("""\
        SELECT CASE sum(blks_hit+blks_read)
          WHEN 0 THEN NULL
          ELSE trunc(sum(blks_hit)/sum(blks_hit+blks_read)*100)::float
        END AS hitratio
        FROM pg_stat_database
        """)

    def get_active_backends(self,):
        current_time = time.time()
        current_active_backends = self._get_current_active_backends()
        return {'nb': current_active_backends,
                'time': current_time}

    def get_cpu_usage(self,):
        sysinfo = SysInfo()
        if sysinfo.os == 'Linux':
            return self._get_cpu_usage_linux()

    def get_load_average(self,):
        return os.getloadavg()[0]

    def get_memory_usage(self,):
        sysinfo = SysInfo()
        if sysinfo.os == 'Linux':
            return self._get_memory_usage_linux()

    def get_stat_db(self,):
        row = self.conn.queryone("""\
        SELECT
            count(datid) as databases,
            pg_size_pretty(sum(pg_database_size(
                pg_database.datname))::bigint) as total_size,
            to_char(now(),'HH24:MI') as time,
            sum(xact_commit)::BIGINT as total_commit,
            sum(xact_rollback)::BIGINT as total_rollback
        FROM pg_database
        JOIN pg_stat_database ON (pg_database.oid = pg_stat_database.datid)
        WHERE datistemplate = 'f'
        """)
        return {'databases': row['databases'],
                'total_size': row['total_size'],
                'time': row['time'],
                'total_commit': row['total_commit'],
                'total_rollback': row['total_rollback'],
                'timestamp': time.time()}

    def get_pg_start_time(self):
        return self.conn.queryscalar("SELECT pg_postmaster_start_time();")

    def _get_memory_usage_linux(self,):
        mem_total = 0
        mem_cached = 0
        mem_free = 0
        mem_active = 0
        pattern_line_meminfo = re.compile(r'^([^:]+):\s+([0-9]+) kB$')
        with open('/proc/meminfo') as fd:
            for line in fd.readlines():
                m = pattern_line_meminfo.match(line)
                if m:
                    key = m.group(1)
                    value = m.group(2)
                    if key == "MemTotal":
                        mem_total = int(value)
                    if key == "MemFree":
                        mem_free = int(value)
                    if key == "Cached":
                        mem_cached = int(value)
        if mem_total == 0:
            raise Exception("Can't parse /proc/meminfo.")
        mem_active = mem_total - mem_free - mem_cached
        return {'total': mem_total,
                'free': round(float(mem_free) / float(mem_total) * 100, 1),
                'active': round(float(mem_active) / float(mem_total) * 100, 1),
                'cached': round(float(mem_cached) / float(mem_total) * 100, 1)}

    def _get_cpu_usage_linux(self,):
        cpu_time_snap_0 = self._get_current_cpu_usage_linux()
        time.sleep(0.1)
        cpu_time_snap_1 = self._get_current_cpu_usage_linux()
        delta_time_total = 0
        delta = {}
        for k in ['time_user', 'time_system', 'time_idle', 'time_iowait',
                  'time_steal']:
            delta_time_total += cpu_time_snap_1[k] - cpu_time_snap_0[k]
            delta[k] = cpu_time_snap_1[k] - cpu_time_snap_0[k]
        return {
            'user': 0 if not delta_time_total else
            round(delta['time_user'] / delta_time_total * 100, 1),
            'system': 0 if not delta_time_total else
            round(delta['time_system'] / delta_time_total * 100, 1),
            'idle': 0 if not delta_time_total else
            round(delta['time_idle'] / delta_time_total * 100, 1),
            'iowait': 0 if not delta_time_total else
            round(delta['time_iowait'] / delta_time_total * 100, 1),
            'steal': 0 if not delta_time_total else
            round(delta['time_steal'] / delta_time_total * 100, 1)
        }

    def _get_current_cpu_usage_linux(self,):
        ret = {}
        with open('/proc/stat') as fd:
            for line in fd.readlines():
                cols = line.split()
                if len(cols) > 0 and cols[0] == 'cpu':
                    ret['time_user'] = float(cols[1]) + float(cols[2])
                    ret['time_system'] = float(cols[3]) + float(cols[6]) + \
                        float(cols[7])
                    ret['time_idle'] = float(cols[4])
                    ret['time_iowait'] = float(cols[5])
                    ret['time_steal'] = float(cols[8])
                    break
        return ret

    def _get_current_buffers(self,):
        return self.conn.queryscalar(
            "SELECT buffers_alloc FROM pg_stat_bgwriter"
        )

    def _get_current_active_backends(self,):
        if self.conn.server_version >= 90200:
            query = """
SELECT COUNT(*) AS nb FROM pg_stat_activity WHERE state != 'idle'
            """
        else:
            query = """
SELECT COUNT(*) AS nb FROM pg_stat_activity WHERE current_query != '<IDLE>'
            """
        return self.conn.queryscalar(query)

    def get_notifications(self, config):
        return list(NotificationMgmt.get_last_n(config, 15))



# Declare and generates openmetrics sample from monitoring data stored in
# SQLite database.
#
# If possible, implement postgres_exporter and node_exporter metrics. Unlike
# prometheus, temBoard agent compute rates and diff in agent side. The open
# metrics must conform to prometheus exporter way : expose raw data (either
# gauge or counter) and let Prometheus query do computations.

import logging
import sys
from typing import List

from ...tools import fromisoformat


logger = logging.getLogger(__name__)
METADATAS = {}


class MetricMetadata:
    COUNTER = 'counter'
    GAUGE = 'gauge'

    def __init__(self, name, help_, type_='untyped'):
        self.name = name
        self.help_ = help_
        self.type_ = type_

    def format(self) -> List[str]:
        yield "# HELP {self.name} {self.help_}".format(self=self)
        yield "# TYPE {self.name} {self.type_}".format(self=self)


class Sample:
    def __init__(self, name, labels=None, value=1, timestamp=None):
        self.name = name
        self.labels = labels or {}
        self.value = value
        self.timestamp = timestamp

    def format(self) -> List[str]:
        labels = ','.join([
            '%s="%s"' % (name, value)
            for name, value in
            sorted(self.labels.items())
        ])
        if labels:
            labels = "{%s}" % labels
        if type(self.value) is str:
            raise Exception(
                "String value %s for %s." % (self.value, self.name))
        elif type(self.value) in (int, float):
            value = str(self.value)
        elif hasattr(self.value, 'timestamp'):
            value = str(self.value.timestamp())
        else:
            raise ValueError("Bad value {self.value}".format(self=self))

        tokens = [
            "{self.name}{labels}".format(self=self, labels=labels),
            value,
            str(self.timestamp or ""),
        ]
        yield " ".join(filter(None, tokens))

    def sort_key(self):
        return (self.name, sorted(self.labels.items()))


def format_open_metrics_lines(samples: List[Sample]) -> List[str]:
    described = set()
    for sample in sorted(samples, key=Sample.sort_key):
        if sample.name not in described:
            if described:
                yield ""
            yield from METADATAS[sample.name].format()
            described.add(sample.name)

        yield from sample.format()

    yield ""
    yield "# EOF"
    yield ""


METADATAS.update({m.name: m for m in [
    MetricMetadata(
        'node_os_info',
        help_=(
            "Operating system metadata, (see node_uname_info for kernel "
            "metadata)."),
        type_=MetricMetadata.GAUGE,
    ),
    MetricMetadata(
        'node_uname_info',
        help_=(
            "Labeled system information as provided by the uname system "
            "call."),
        type_=MetricMetadata.GAUGE,
    ),
    MetricMetadata(
        'pg_settings_max_connections',
        help_="Sets the maximum number of concurrent connections.",
        type_=MetricMetadata.GAUGE,
    ),
    # For compatibility with postgres_exporter.
    MetricMetadata(
        'pg_exporter_last_scrape_duration_seconds',
        help_="Duration of the last scrape of metrics from PostgresSQL.",
        type_=MetricMetadata.GAUGE,
    ),
    MetricMetadata(
        'pg_static',
        help_="Version string as reported by postgres",
    ),
    MetricMetadata(
        'pg_up',
        help_=(
            "Whether the last scrape of metrics from PostgreSQL was able to "
            "connect to the server (1 for yes, 0 for no)."),
        type_=MetricMetadata.GAUGE,
    ),
    # From
    # https://github.com/prometheus-community/postgres_exporter/blob/master/queries.yaml#L9-L15
    # Required by PostgreSQL Database dashboard variables.
    MetricMetadata(
        'pg_postmaster_start_time_seconds',
        help_="Time at which postmaster started",
        type_=MetricMetadata.GAUGE,
    ),
]})


def generate_samples(temboard_data) -> List[Sample]:
    instance = temboard_data['instances'][0]
    yield Sample(
        'node_os_info',
        dict(pretty_name=temboard_data['hostinfo']['os_flavor']),
    )
    nodename, _, domainname = instance['hostname'].partition('.')
    yield Sample(
        'node_uname_info',
        dict(
            domainename=domainname or '(none)',
            machine=temboard_data['hostinfo']['cpu_arch'],
            nodename=nodename,
            release=temboard_data['hostinfo']['os_version'],
            sysname=temboard_data['hostinfo']['os'],
        ),
    )

    yield Sample('pg_up')
    yield Sample(
        'pg_exporter_last_scrape_duration_seconds',
        value=0.0
    )
    yield Sample(
        'pg_postmaster_start_time_seconds',
        value=fromisoformat(instance['start_time']),
    )
    yield Sample(
        'pg_settings_max_connections',
        value=int(instance['max_connections']),
    )
    yield Sample(
        'pg_static',
        dict(
            short_version=instance['version'],
            # Extension of postgres_exporter labels.
            temboard_agent_version=temboard_data['version'],
            cluster_name=instance['local_name'],
        ),
    )

    mod = sys.modules[__name__]
    for probe, data in temboard_data['data'].items():
        if not data:
            continue

        generator_name = 'generate_%s_samples' % probe
        generator = getattr(mod, generator_name, None)
        if not generator:
            logger.debug("Can't export metrics from probe %s.", probe)
            raise Exception(probe)
            continue

        yield from generator(data, temboard_data['hostinfo'])


METADATAS.update({m.name: m for m in [
    MetricMetadata(
        'node_load1',
        help_="1m load average.",
        type_=MetricMetadata.GAUGE,
    ),
    MetricMetadata(
        'node_load5',
        help_="5m load average.",
        type_=MetricMetadata.GAUGE,
    ),
    MetricMetadata(
        'node_load15',
        help_="15m load average.",
        type_=MetricMetadata.GAUGE,
    ),
]})


def generate_loadavg_samples(data, hostinfo):
    for relative in '1', '5', '15':
        yield Sample(
            'node_load' + relative,
            value=float(data[0]['load' + relative]),
        )


METADATAS.update({m.name: m for m in [
    MetricMetadata(
        'node_cpu_seconds_total',
        help_="Seconds the CPUs spent in each mode.",
        type_=MetricMetadata.COUNTER,
    ),
]})


def generate_cpu_samples(data, hostinfo):
    for mode in 'idle', 'iowait', 'steal', 'system', 'user':
        for core in range(hostinfo['cpu_count']):
            yield Sample(
                'node_cpu_seconds_total',
                dict(cpu=core, mode=mode),
                # temBoard use global cpu data while node_exporter use per-core
                # data. Divide by core count to mimic per-core data.
                data[0]['time_' + mode] / 1000 / hostinfo['cpu_count'],
            )


METADATAS.update({m.name: m for m in [
    *[MetricMetadata(
        'node_memory_%s_bytes' % field,
        help_="Memory information field %s_bytes." % field,
        type_=MetricMetadata.GAUGE,
    ) for field in (
        'Buffers', 'Cached',
        'MemFree', 'MemTotal',
        'SwapTotal', 'SwapFree',
    )],
]})


def generate_memory_samples(data, hostinfo):
    yield Sample(
        'node_memory_Buffers_bytes',
        value=data[0]['mem_buffers'],
    )
    yield Sample(
        'node_memory_Cached_bytes',
        value=data[0]['mem_cached'],
    )
    yield Sample(
        'node_memory_MemFree_bytes',
        value=data[0]['mem_free'],
    )
    yield Sample(
        'node_memory_MemTotal_bytes',
        value=data[0]['mem_total'],
    )
    yield Sample(
        'node_memory_SwapFree_bytes',
        value=data[0]['swap_total'] - data[0]['swap_used'],
    )
    yield Sample(
        'node_memory_SwapTotal_bytes',
        value=data[0]['swap_total'],
    )


METADATAS.update({m.name: m for m in [
    MetricMetadata(
        'node_context_switches_total',
        help_="Total number of context switches.",
        type_=MetricMetadata.COUNTER,
    ),
    MetricMetadata(
        'node_fork_total',
        help_="Total number of forks.",
        type_=MetricMetadata.COUNTER,
    ),
    MetricMetadata(
        'node_procs_blocked',
        help_="Number of processes blocked waiting for I/O to complete.",
        type_=MetricMetadata.GAUGE,
    ),
    MetricMetadata(
        'node_procs_running',
        help_="Number of processes in runnable state.",
        type_=MetricMetadata.GAUGE,
    ),
    MetricMetadata(
        'xnode_procs_total',
        help_="Total number of processes (temBoard custom).",
        type_=MetricMetadata.GAUGE,
    ),
]})


def generate_process_samples(data, hostinfo):
    yield Sample(
        'node_context_switches_total',
        value=data[0]['context_switches'],
    )
    yield Sample(
        'node_fork_total',
        value=data[0]['forks'],
    )
    for field in 'blocked', 'running':
        yield Sample(
            'node_procs_' + field,
            value=data[0]['procs_' + field],
        )
    yield Sample(
        'xnode_procs_total',
        value=int(data[0]['procs_total']),
    )


METADATAS.update({m.name: m for m in [
    MetricMetadata(
        'node_filesystem_size_bytes',
        help_="Filesystem size in bytes.",
        type_=MetricMetadata.GAUGE,
    ),
    MetricMetadata(
        'node_filesystem_free_bytes',
        help_="Filesystem free space in bytes.",
        type_=MetricMetadata.GAUGE,
    ),
]})


def generate_filesystems_size_samples(data, hostinfo):
    for fs in data:
        yield Sample(
            'node_filesystem_size_bytes',
            dict(device=fs['device'], mountpoint=fs['mount_point']),
            fs['total'],
        )
        yield Sample(
            'node_filesystem_free_bytes',
            dict(device=fs['device'], mountpoint=fs['mount_point']),
            fs['total'] - fs['used'],
        )


METADATAS.update({m.name: m for m in [
    MetricMetadata(
        'pg_locks_count',
        help_="Number of locks",
        type_=MetricMetadata.GAUGE,
    ),
]})


def generate_locks_samples(data, hostinfo):
    lock_modes = (
        'access_share',
        'waiting_access_share',
        'row_share',
        'waiting_row_share',
        'row_exclusive',
        'waiting_row_exclusive',
        'share_update_exclusive',
        'waiting_share_update_exclusive',
        'share',
        'waiting_share',
        'share_row_exclusive',
        'waiting_share_row_exclusive',
        'exclusive',
        'waiting_exclusive',
        'access_exclusive',
        'waiting_access_exclusive',
        'siread',
    )
    for datdata in data:
        for mode in lock_modes:
            prommode = mode.replace('_', '')
            yield Sample(
                'pg_locks_count',
                dict(mode=prommode, datname=datdata['dbname']),
                datdata[mode],
            )


METADATAS.update({m.name: m for m in [
    MetricMetadata(
        'pg_stat_activity_count',
        help_="number of connections in this state",
        type_=MetricMetadata.GAUGE,
    ),
]})


def generate_sessions_samples(data, hostinfo):
    session_states = {
        'active': 'active',
        'disabled': 'disabled',
        'fastpath': 'fastpath function call',
        'idle': 'idle',
        'idle_in_xact': 'idle in transaction',
        'idle_in_xact_aborted': 'idle in transaction (aborted)',
    }
    for datdata in data:
        for state, promstate in session_states.items():
            yield Sample(
                'pg_stat_activity_count',
                dict(datname=datdata['dbname'], state=promstate),
                datdata[state]
            )


METADATAS.update({m.name: m for m in [
    MetricMetadata(
        'pg_stat_bgwriter_buffers_alloc',
        help_="Number of buffers allocated",
        type_=MetricMetadata.COUNTER,
    ),
    MetricMetadata(
        'pg_stat_bgwriter_buffers_backend',
        help_="Number of buffers written directly by a backend",
        type_=MetricMetadata.COUNTER,
    ),
    MetricMetadata(
        'pg_stat_bgwriter_buffers_backend_fsync',
        help_=(
            "Number of times a backend had to execute its own fsync call "
            "(normally the background writer handles those even when the "
            "backend does its own write)"
        ),
        type_=MetricMetadata.COUNTER,
    ),
    MetricMetadata(
        'pg_stat_bgwriter_buffers_checkpoint',
        help_="Number of buffers written during checkpoints",
        type_=MetricMetadata.COUNTER,
    ),
    MetricMetadata(
        'pg_stat_bgwriter_buffers_clean',
        help_="Number of buffers written by the background writer",
        type_=MetricMetadata.COUNTER,
    ),
    MetricMetadata(
        'pg_stat_bgwriter_checkpoint_sync_time',
        help_=(
            "Total amount of time that has been spent in the portion of "
            "checkpoint processing where files are synchronized to disk, "
            "in milliseconds"
        ),
        type_=MetricMetadata.COUNTER,
    ),
    MetricMetadata(
        'pg_stat_bgwriter_checkpoint_write_time',
        help_=(
            "Total amount of time that has been spent in the portion of "
            "checkpoint processing where files are written to disk, in "
            "milliseconds"
        ),
        type_=MetricMetadata.COUNTER,
    ),
    MetricMetadata(
        'pg_stat_bgwriter_checkpoints_req',
        help_="Number of requested checkpoints that have been performed",
        type_=MetricMetadata.COUNTER,
    ),
    MetricMetadata(
        'pg_stat_bgwriter_checkpoints_timed',
        help_="Number of scheduled checkpoints that have been performed",
        type_=MetricMetadata.COUNTER,
    ),
    MetricMetadata(
        'pg_stat_bgwriter_maxwritten_clean',
        help_=(
            "Number of times the background writer stopped a cleaning scan "
            "because it had written too many buffers"
        ),
        type_=MetricMetadata.COUNTER,
    ),
    MetricMetadata(
        'pg_stat_bgwriter_stats_reset',
        help_="Time at which these statistics were last reset",
        type_=MetricMetadata.COUNTER,
    ),
]})


def generate_bgwriter_samples(data, hostinfo):
    yield Sample(
        'pg_stat_bgwriter_stats_reset',
        value=fromisoformat(data[0]['stats_reset']),
    )

    bgwriter_stats = (
        'buffers_alloc',
        'buffers_backend',
        'buffers_backend_fsync',
        'buffers_clean',
        'buffers_checkpoint',
        'checkpoint_sync_time',
        'checkpoint_write_time',
        'checkpoints_timed',
        'checkpoints_req',
        'maxwritten_clean',
    )
    for stat in bgwriter_stats:
        yield Sample(
            'pg_stat_bgwriter_' + stat,
            value=data[0][stat],
        )


METADATAS.update({m.name: m for m in [
    MetricMetadata(
        'xpg_database_size_bytes',
        help_="Database total size",
        type_=MetricMetadata.GAUGE,
    ),
]})


def generate_db_size_samples(data, hostinfo):
    for datdata in data:
        yield Sample(
            'xpg_database_size_bytes',
            dict(datname=datdata['dbname']),
            datdata['size'],
        )


METADATAS.update({m.name: m for m in [
    MetricMetadata(
        'pg_stat_database_temp_bytes',
        help_=(
            "Total amount of data written to temporary files by queries in "
            "this database. All temporary files are counted, regardless of "
            "why the temporary file was created, and regardless of the "
            "log_temp_files setting.",
        ),
        type_=MetricMetadata.COUNTER,
    ),
]})


def generate_temp_files_size_delta_samples(data, hostinfo):
    for datdata in data:
        yield Sample(
            'pg_stat_database_temp_bytes',
            dict(datname=datdata['dbname']),
            datdata['size'],
        )


METADATAS.update({m.name: m for m in [
    MetricMetadata(
        'xpg_tablespace_size_bytes',
        help_="Tablespace total size",
        type_=MetricMetadata.GAUGE,
    ),
]})


def generate_tblspc_size_samples(data, hostinfo):
    for spcdata in data:
        yield Sample(
            'xpg_tablespace_size_bytes',
            dict(spcname=spcdata['spcname']),
            spcdata['size']
        )


METADATAS.update({m.name: m for m in [
    MetricMetadata(
        'xpg_wal_files_bytes_total',
        help_="Total size of all WALs",
        type_=MetricMetadata.GAUGE,
    ),
    MetricMetadata(
        'xpg_wal_files_total',
        help_="Total number of WALs",
        type_=MetricMetadata.GAUGE,
    ),
    MetricMetadata(
        'xpg_wal_files_archive_ready_total',
        help_="Total number of WALs ready for archiving.",
        type_=MetricMetadata.GAUGE,
    ),
    MetricMetadata(
        'xpg_wal_files_written_size_bytes',
        help_="Total bytes written in current WAL.",
        type_=MetricMetadata.COUNTER,
    ),
]})


def generate_wal_files_samples(data, hostinfo):
    yield Sample(
        'xpg_wal_files_total',
        value=data[0]['total'],
    )
    yield Sample(
        'xpg_wal_files_archive_ready_total',
        value=data[0]['archive_ready'],
    )
    yield Sample(
        'xpg_wal_files_bytes_total',
        value=data[0]['total_size'],
    )
    yield Sample(
        'xpg_wal_files_written_size_bytes',
        dict(current_wal_lsn=data[0]['current_location']),
        data[0]['written_size'],
    )


METADATAS.update({m.name: m for m in [
    MetricMetadata(
        'xpg_btree_bloat_ratio',
        help_="Bloat estimation of btree indexes in database.",
        type_=MetricMetadata.GAUGE,
    ),
]})


def generate_btree_bloat_samples(data, hostinfo):
    for datdata in data:
        yield Sample(
            'xpg_btree_bloat_ratio',
            dict(datname=datdata['dbname']),
            datdata['ratio'],
        )


METADATAS.update({m.name: m for m in [
    MetricMetadata(
        'xpg_heap_bloat_ratio',
        help_="Bloat estimation of table data in database.",
        type_=MetricMetadata.GAUGE,
    ),
]})


def generate_heap_bloat_samples(data, hostinfo):
    for datdata in data:
        yield Sample(
            'xpg_heap_bloat_ratio',
            dict(datname=datdata['dbname']),
            datdata['ratio'],
        )


METADATAS.update({m.name: m for m in [
    MetricMetadata(
        'pg_stat_database_xact_commit',
        help_=(
            "Number of transactions in this database that have been "
            "committed"
        ),
        type_=MetricMetadata.COUNTER,
    ),
    MetricMetadata(
        'pg_stat_database_xact_rollback',
        help_=(
            "Number of transactions in this database that have been rolled "
            "back"
        ),
        type_=MetricMetadata.COUNTER,
    ),
]})


def generate_xacts_samples(data, hostinfo):
    for datdata in data:
        yield Sample(
            'pg_stat_database_xact_commit',
            dict(datname=datdata['dbname']),
            datdata['n_commit'],
        )
        yield Sample(
            'pg_stat_database_xact_rollback',
            dict(datname=datdata['dbname']),
            datdata['n_rollback'],
        )


METADATAS.update({m.name: m for m in [
    MetricMetadata(
        'pg_stat_database_blks_hit',
        help_=(
            "Number of times disk blocks were found already in the buffer "
            "cache, so that a read was not necessary (this only includes hits "
            "in the PostgreSQL buffer cache, not the operating system's file "
            "system cache)"
        ),
        type_=MetricMetadata.COUNTER,
    ),
    MetricMetadata(
        'pg_stat_database_blks_read',
        help_="Number of disk blocks read in this database",
        type_=MetricMetadata.COUNTER,
    ),
]})


def generate_blocks_samples(data, hostinfo):
    for datdata in data:
        yield Sample(
            'pg_stat_database_blks_hit',
            dict(datname=datdata['dbname']),
            datdata['blks_hit'],
        )
        yield Sample(
            'pg_stat_database_blks_read',
            dict(datname=datdata['dbname']),
            datdata['blks_read'],
        )
        # let PromQL compute hitmiss ratio.


METADATAS.update({m.name: m for m in [
    MetricMetadata(
        'xpg_replication_up',
        help_="Streaming replication connection status",
        type_=MetricMetadata.GAUGE,
    ),
]})


def generate_replication_connection_samples(data, hostinfo):
    yield Sample(
        'xpg_replication_up',
        dict(upstream=data[0]['upstream']),
        data[0]['connected'],
    )


METADATAS.update({m.name: m for m in [
    MetricMetadata(
        'xpg_replication_lag_bytes',
        help_="Streaming replication lag",
        type_=MetricMetadata.GAUGE,
    ),
]})


def generate_replication_lag_samples(data, hostinfo):
    yield Sample(
        'xpg_replication_lag_bytes',
        value=data[0]['lag'],
    )



from datetime import datetime, timedelta
import hashlib
import logging
import os

from bottle import HTTPError

from temboardagent.errors import UserError
from temboardagent.toolkit import taskmanager

logger = logging.getLogger(__package__)

# Taken from https://github.com/ioguix/pgsql-bloat-estimation/blob/master/table/table_bloat.sql  # noqa
TABLE_BLOAT_SQL = """
SELECT current_database(), schemaname, tblname, bs*tblpages AS real_size,
  (tblpages-est_tblpages)*bs AS extra_size,
  CASE WHEN tblpages - est_tblpages > 0
    THEN 100 * (tblpages - est_tblpages)/tblpages::float
    ELSE 0
  END AS extra_ratio, fillfactor,
  CASE WHEN tblpages - est_tblpages_ff > 0
    THEN (tblpages-est_tblpages_ff)*bs
    ELSE 0
  END AS bloat_size,
  CASE WHEN tblpages - est_tblpages_ff > 0
    THEN 100 * (tblpages - est_tblpages_ff)/tblpages::float
    ELSE 0
  END AS bloat_ratio, is_na
  -- , (pst).free_percent + (pst).dead_tuple_percent AS real_frag
FROM (
  SELECT ceil( reltuples / ( (bs-page_hdr)/tpl_size ) ) + ceil( toasttuples / 4 ) AS est_tblpages,
    ceil( reltuples / ( (bs-page_hdr)*fillfactor/(tpl_size*100) ) ) + ceil( toasttuples / 4 ) AS est_tblpages_ff,
    tblpages, fillfactor, bs, tblid, schemaname, tblname, heappages, toastpages, is_na
    -- , stattuple.pgstattuple(tblid) AS pst
  FROM (
    SELECT
      ( 4 + tpl_hdr_size + tpl_data_size + (2*ma)
        - CASE WHEN tpl_hdr_size%ma = 0 THEN ma ELSE tpl_hdr_size%ma END
        - CASE WHEN ceil(tpl_data_size)::int%ma = 0 THEN ma ELSE ceil(tpl_data_size)::int%ma END
      ) AS tpl_size, bs - page_hdr AS size_per_block, (heappages + toastpages) AS tblpages, heappages,
      toastpages, reltuples, toasttuples, bs, page_hdr, tblid, schemaname, tblname, fillfactor, is_na
    FROM (
      SELECT
        tbl.oid AS tblid, ns.nspname AS schemaname, tbl.relname AS tblname, tbl.reltuples,
        tbl.relpages AS heappages, coalesce(toast.relpages, 0) AS toastpages,
        coalesce(toast.reltuples, 0) AS toasttuples,
        coalesce(substring(
          array_to_string(tbl.reloptions, ' ')
          FROM 'fillfactor=([0-9]+)')::smallint, 100) AS fillfactor,
        current_setting('block_size')::numeric AS bs,
        CASE WHEN version()~'mingw32' OR version()~'64-bit|x86_64|ppc64|ia64|amd64' THEN 8 ELSE 4 END AS ma,
        24 AS page_hdr,
        23 + CASE WHEN MAX(coalesce(s.null_frac,0)) > 0 THEN ( 7 + count(s.attname) ) / 8 ELSE 0::int END
           + CASE WHEN bool_or(att.attname = 'oid' and att.attnum < 0) THEN 4 ELSE 0 END AS tpl_hdr_size,
        sum( (1-coalesce(s.null_frac, 0)) * coalesce(s.avg_width, 0) ) AS tpl_data_size,
        bool_or(att.atttypid = 'pg_catalog.name'::regtype)
          OR sum(CASE WHEN att.attnum > 0 THEN 1 ELSE 0 END) <> count(s.attname) AS is_na
      FROM pg_attribute AS att
        JOIN pg_class AS tbl ON att.attrelid = tbl.oid
        JOIN pg_namespace AS ns ON ns.oid = tbl.relnamespace
        LEFT JOIN pg_stats AS s ON s.schemaname=ns.nspname
          AND s.tablename = tbl.relname AND s.inherited=false AND s.attname=att.attname
        LEFT JOIN pg_class AS toast ON tbl.reltoastrelid = toast.oid
      WHERE NOT att.attisdropped
        AND tbl.relkind = 'r'
      GROUP BY 1,2,3,4,5,6,7,8,9,10
      ORDER BY 2,3
    ) AS s
  ) AS s2
) AS s3
"""  # noqa


INDEX_BTREE_BLOAT_SQL = """
-- This query must be exected by a superuser because it relies on the
-- pg_statistic table.
-- This query run much faster than btree_bloat.sql, about 1000x faster.
--
-- This query is compatible with PostgreSQL 8.2 and after.
SELECT current_database(), nspname AS schemaname, tblname, idxname, bs*(relpages)::bigint AS real_size,
  bs*(relpages-est_pages)::bigint AS extra_size,
  100 * (relpages-est_pages)::float / relpages AS extra_ratio,
  fillfactor,
  CASE WHEN relpages > est_pages_ff
    THEN bs*(relpages-est_pages_ff)
    ELSE 0
  END AS bloat_size,
  100 * (relpages-est_pages_ff)::float / relpages AS bloat_ratio,
  is_na
  -- , 100-(pst).avg_leaf_density AS pst_avg_bloat, est_pages, index_tuple_hdr_bm, maxalign, pagehdr, nulldatawidth, nulldatahdrwidth, reltuples, relpages -- (DEBUG INFO)
FROM (
  SELECT coalesce(1 +
         ceil(reltuples/floor((bs-pageopqdata-pagehdr)/(4+nulldatahdrwidth)::float)), 0 -- ItemIdData size + computed avg size of a tuple (nulldatahdrwidth)
      ) AS est_pages,
      coalesce(1 +
         ceil(reltuples/floor((bs-pageopqdata-pagehdr)*fillfactor/(100*(4+nulldatahdrwidth)::float))), 0
      ) AS est_pages_ff,
      bs, nspname, tblname, idxname, relpages, fillfactor, is_na
      -- , pgstatindex(idxoid) AS pst, index_tuple_hdr_bm, maxalign, pagehdr, nulldatawidth, nulldatahdrwidth, reltuples -- (DEBUG INFO)
  FROM (
      SELECT maxalign, bs, nspname, tblname, idxname, reltuples, relpages, idxoid, fillfactor,
            ( index_tuple_hdr_bm +
                maxalign - CASE -- Add padding to the index tuple header to align on MAXALIGN
                  WHEN index_tuple_hdr_bm%maxalign = 0 THEN maxalign
                  ELSE index_tuple_hdr_bm%maxalign
                END
              + nulldatawidth + maxalign - CASE -- Add padding to the data to align on MAXALIGN
                  WHEN nulldatawidth = 0 THEN 0
                  WHEN nulldatawidth::integer%maxalign = 0 THEN maxalign
                  ELSE nulldatawidth::integer%maxalign
                END
            )::numeric AS nulldatahdrwidth, pagehdr, pageopqdata, is_na
            -- , index_tuple_hdr_bm, nulldatawidth -- (DEBUG INFO)
      FROM (
          SELECT n.nspname, ct.relname AS tblname, i.idxname, i.reltuples, i.relpages,
              i.idxoid, i.fillfactor, current_setting('block_size')::numeric AS bs,
              CASE -- MAXALIGN: 4 on 32bits, 8 on 64bits (and mingw32 ?)
                WHEN version() ~ 'mingw32' OR version() ~ '64-bit|x86_64|ppc64|ia64|amd64' THEN 8
                ELSE 4
              END AS maxalign,
              /* per page header, fixed size: 20 for 7.X, 24 for others */
              24 AS pagehdr,
              /* per page btree opaque data */
              16 AS pageopqdata,
              /* per tuple header: add IndexAttributeBitMapData if some cols are null-able */
              CASE WHEN max(coalesce(s.stanullfrac,0)) = 0
                  THEN 2 -- IndexTupleData size
                  ELSE 2 + (( 32 + 8 - 1 ) / 8) -- IndexTupleData size + IndexAttributeBitMapData size ( max num filed per index + 8 - 1 /8)
              END AS index_tuple_hdr_bm,
              /* data len: we remove null values save space using it fractionnal part from stats */
              sum( (1-coalesce(s.stanullfrac, 0)) * coalesce(s.stawidth, 1024)) AS nulldatawidth,
              max( CASE WHEN a.atttypid = 'pg_catalog.name'::regtype THEN 1 ELSE 0 END ) > 0 AS is_na
          FROM (
              SELECT idxname, reltuples, relpages, tbloid, idxoid, fillfactor,
                  CASE WHEN indkey[i]=0 THEN idxoid ELSE tbloid END AS att_rel,
                  CASE WHEN indkey[i]=0 THEN i ELSE indkey[i] END AS att_pos
              FROM (
                  SELECT idxname, reltuples, relpages, tbloid, idxoid, fillfactor, indkey, generate_series(1,indnatts) AS i
                  FROM (
                      SELECT ci.relname AS idxname, ci.reltuples, ci.relpages, i.indrelid AS tbloid,
                          i.indexrelid AS idxoid,
                          coalesce(substring(
                              array_to_string(ci.reloptions, ' ')
                              from 'fillfactor=([0-9]+)')::smallint, 90) AS fillfactor,
                          i.indnatts,
                          string_to_array(textin(int2vectorout(i.indkey)),' ')::int[] AS indkey
                      FROM pg_index i
                      JOIN pg_class ci ON ci.oid=i.indexrelid
                      WHERE ci.relam=(SELECT oid FROM pg_am WHERE amname = 'btree')
                        AND ci.relpages > 0
                  ) AS idx_data
              ) AS idx_data_cross
          ) i
          JOIN pg_attribute a ON a.attrelid = i.att_rel
                             AND a.attnum = i.att_pos
          JOIN pg_statistic s ON s.starelid = i.att_rel
                             AND s.staattnum = i.att_pos
          JOIN pg_class ct ON ct.oid = i.tbloid
          JOIN pg_namespace n ON ct.relnamespace = n.oid
          GROUP BY 1,2,3,4,5,6,7,8,9,10
      ) AS rows_data_stats
  ) AS rows_hdr_pdg_stats
) AS relation_stats
ORDER BY nspname, tblname, idxname
"""  # noqa


SCHEMAS_SQL = """
SELECT n.nspname AS "name",
       COALESCE(schema_size, 0) AS total_bytes,
       pg_size_pretty(schema_size) AS total_size,
       COALESCE(n_tables, 0) AS n_tables,
       COALESCE(tables_bytes, 0) AS tables_bytes,
       pg_size_pretty(tables_bytes) AS tables_size,
       COALESCE(n_indexes, 0) AS n_indexes,
       COALESCE(indexes_bytes, 0) AS indexes_bytes,
       pg_size_pretty(indexes_bytes) AS indexes_size,
       tbloat.bloat_size AS tables_bloat_bytes,
       pg_size_pretty(tbloat.bloat_size::bigint) AS tables_bloat_size,
       ibloat.bloat_size AS indexes_bloat_bytes,
       pg_size_pretty(ibloat.bloat_size::bigint) AS indexes_bloat_size,
       COALESCE(toast.toast_bytes, 0) AS toast_bytes,
       pg_size_pretty(toast.toast_bytes) AS toast_size
FROM pg_catalog.pg_namespace n
-- schema size + tables for the schema (count, size)
-- See https://wiki.postgresql.org/wiki/Schema_Size
LEFT JOIN (
  SELECT schemaname,
         SUM(pg_total_relation_size(quote_ident(schemaname) || '.' || quote_ident(tablename)))::BIGINT AS schema_size,
         count(*) as n_tables,
         SUM(pg_relation_size(quote_ident(schemaname) || '.' || quote_ident(tablename)))::BIGINT AS tables_bytes
  FROM pg_tables
  GROUP BY schemaname
) AS a
ON a.schemaname = n.nspname
-- toast size
LEFT JOIN (
  SELECT nspname,
         SUM(pg_total_relation_size(reltoastrelid)) AS toast_bytes
  FROM pg_class c
  LEFT JOIN pg_namespace n ON n.oid = c.relnamespace
  GROUP BY nspname
) AS toast
ON toast.nspname = n.nspname
-- indexes for the schema (count, size)
LEFT JOIN (
  SELECT count(*) as n_indexes,
         schemaname,
         SUM(pg_relation_size(quote_ident(schemaname) || '.' || quote_ident(indexname)))::BIGINT AS indexes_bytes
  FROM pg_catalog.pg_indexes
  GROUP BY schemaname
) AS indexes
ON indexes.schemaname = n.nspname
LEFT JOIN (
  SELECT SUM(bloat_size) AS bloat_size,
         schemaname
  FROM (
    %s
  ) AS a
  GROUP BY schemaname
) AS tbloat
ON tbloat.schemaname = n.nspname
LEFT JOIN (
  SELECT SUM(bloat_size) AS bloat_size,
         schemaname
  FROM (
    %s
  ) AS a
  GROUP BY schemaname
) AS ibloat
ON ibloat.schemaname = n.nspname
WHERE n.nspname !~ '^pg_temp'
AND n.nspname !~ '^pg_toast'
""" % (TABLE_BLOAT_SQL, INDEX_BTREE_BLOAT_SQL)  # noqa


INDEXES_SQL = """
SELECT i.tablename AS tablename,
       i.indexname AS name,
       tablespace,
       x.indnatts AS number_of_columns,
       idx_scan AS scans,
       idx_tup_read,
       idx_tup_fetch,
       indexrelname,
       indisunique,
       i.indexdef AS def,
       total_bytes,
       pg_size_pretty(total_bytes) AS total_size,
       am.amname AS type,
       ibloat.bloat_size AS bloat_bytes,
       pg_size_pretty(ibloat.bloat_size::bigint) AS bloat_size
FROM pg_index x
JOIN (
    select oid, relname, relam, pg_total_relation_size(c.oid) AS total_bytes
    FROM pg_class c
) AS c
ON c.oid = x.indexrelid
JOIN pg_catalog.pg_indexes i
ON c.relname = i.indexname
JOIN pg_stat_all_indexes psai
ON x.indexrelid = psai.indexrelid
JOIN pg_am am
ON am.oid = c.relam
JOIN (
    WITH qq AS (
    """ + INDEX_BTREE_BLOAT_SQL + """
    ) SELECT * FROM qq
) AS ibloat
ON ibloat.schemaname = i.schemaname AND ibloat.tblname = i.tablename AND ibloat.idxname = i.indexname
WHERE i.schemaname = '{schema}'
{table_filter}
ORDER BY 1,2
"""  # noqa


def get_instance(conn):
    return conn.query("""\
    SELECT SUM(pg_database_size(datname)) AS total_bytes,
        pg_size_pretty(SUM(pg_database_size(datname))) AS total_size
    FROM pg_database
    WHERE NOT datistemplate;
    """)


def get_databases(conn):
    return list(conn.query("""\
    SELECT datname,
        pg_database_size(datname) AS total_bytes,
        pg_size_pretty(pg_database_size(datname)) AS total_size
    FROM pg_database
    WHERE NOT datistemplate;
    """))


def get_database_size(conn):
    return conn.queryone("""\
    SELECT pg_size_pretty(pg_database_size(current_database())) AS total_size,
        pg_database_size(current_database()) AS total_bytes
    """)


def get_database(conn):
    return conn.queryone("""
    SELECT SUM(n_tables) AS n_tables,
        SUM(tables_bytes) as tables_bytes,
        pg_size_pretty(SUM(tables_bytes)) AS tables_size,
        SUM(n_indexes) AS n_indexes,
        SUM(indexes_bytes) AS indexes_bytes,
        pg_size_pretty(SUM(indexes_bytes)) AS indexes_size,
        SUM(tables_bloat_bytes) AS tables_bloat_bytes,
        pg_size_pretty(SUM(tables_bloat_bytes)::bigint) AS tables_bloat_size,
        SUM(indexes_bloat_bytes) AS indexes_bloat_bytes,
        pg_size_pretty(SUM(indexes_bloat_bytes)::bigint) AS indexes_bloat_size,
        SUM(toast_bytes) AS toast_bytes,
        pg_size_pretty(SUM(toast_bytes)::bigint) AS toast_size
    FROM (%s) a""" % SCHEMAS_SQL)


def get_schemas(conn):
    return list(conn.query(SCHEMAS_SQL))


def get_schema(conn, schema):
    rows = conn.query("""\
    SELECT pg_size_pretty(bytes) AS size,  COALESCE(bytes, 0) as total_bytes
    FROM (
        SELECT schemaname, SUM(pg_total_relation_size(quote_ident(schemaname) || '.' || quote_ident(tablename)))::BIGINT AS bytes
        FROM pg_tables
        GROUP BY schemaname
    ) a
    WHERE schemaname = '{schema}'
    """.format(schema=schema)  # noqa
    )
    try:
        return next(rows)
    except StopIteration:
        return {}


def get_tables(conn, schema):
    # taken from https://wiki.postgresql.org/wiki/Disk_Usage
    query = """
SELECT table_name AS name,
       total_bytes,
       index_bytes,
       toast_bytes,
       table_bytes,
       COALESCE(n_indexes, 0) AS n_indexes,
       pg_size_pretty(total_bytes) AS total_size,
       pg_size_pretty(index_bytes) AS index_size,
       pg_size_pretty(toast_bytes) AS toast_size,
       pg_size_pretty(table_bytes) AS table_size,
       tbloat.bloat_size AS bloat_bytes,
       pg_size_pretty(tbloat.bloat_size::bigint) AS bloat_size,
       ibloat.bloat_size AS index_bloat_bytes,
       pg_size_pretty(ibloat.bloat_size::bigint) AS index_bloat_size,
       row_estimate
FROM (
  SELECT *, total_bytes - index_bytes - toast_bytes AS table_bytes
  FROM (
    SELECT c.oid,nspname AS table_schema,
           relname AS TABLE_NAME,
           c.reltuples AS row_estimate,
           pg_total_relation_size(c.oid) AS total_bytes,
           pg_indexes_size(c.oid) AS index_bytes,
           COALESCE(pg_total_relation_size(reltoastrelid), 0) AS toast_bytes
    FROM pg_class c
    LEFT JOIN pg_namespace n ON n.oid = c.relnamespace
    WHERE relkind = 'r'
  ) a
) a
LEFT JOIN (
  SELECT count(*) as n_indexes, tablename
  FROM pg_catalog.pg_indexes
  GROUP BY tablename
) AS indexes
ON indexes.tablename = table_name
JOIN (
  """ + TABLE_BLOAT_SQL + """
) AS tbloat
ON tbloat.schemaname = table_schema AND tbloat.tblname = table_name
LEFT JOIN (
  SELECT SUM(bloat_size) AS bloat_size,
         schemaname,
         tblname
  FROM (
    """ + INDEX_BTREE_BLOAT_SQL + """
  ) AS a
  GROUP BY schemaname, tblname
) AS ibloat
ON ibloat.schemaname = table_schema AND ibloat.tblname = table_name
WHERE table_schema = '{schema}';
    """ # noqa
    return {
        'tables': list(conn.query(query.format(schema=schema)))
    }


def get_schema_indexes(conn, schema):
    return {'indexes': list(conn.query(
        INDEXES_SQL.format(schema=schema, table_filter='')
    ))}


def get_table_indexes(conn, schema, table):
    return {'indexes': list(conn.query(
        INDEXES_SQL.format(
            schema=schema,
            table_filter="AND i.tablename = '%s'" % table
        )
    ))}


def get_table(conn, schema, table):
    query = """
SELECT table_name AS name,
       total_bytes,
       index_bytes,
       toast_bytes,
       table_bytes,
       pg_size_pretty(total_bytes) AS total_size,
       pg_size_pretty(index_bytes) AS index_size,
       pg_size_pretty(toast_bytes) AS toast_size,
       pg_size_pretty(table_bytes) AS table_size,
       pg_stat_all_tables.*,
       tbloat.bloat_size AS bloat_bytes,
       pg_size_pretty(tbloat.bloat_size::bigint) AS bloat_size,
       ibloat.bloat_size AS index_bloat_bytes,
       pg_size_pretty(ibloat.bloat_size::bigint) AS index_bloat_size,
       row_estimate,
       fillfactor
FROM (
  SELECT *, total_bytes - index_bytes - COALESCE(toast_bytes,0) AS table_bytes
  FROM (
    SELECT c.oid,nspname AS table_schema,
           relname AS TABLE_NAME,
           c.reltuples AS row_estimate,
           pg_total_relation_size(c.oid) AS total_bytes,
           pg_indexes_size(c.oid) AS index_bytes,
           pg_total_relation_size(reltoastrelid) AS toast_bytes
    FROM pg_class c
    LEFT JOIN pg_namespace n ON n.oid = c.relnamespace
    WHERE relkind = 'r'
  ) a
) a
JOIN (
  """ + TABLE_BLOAT_SQL + """
) AS tbloat
ON tbloat.schemaname = table_schema AND tbloat.tblname = table_name
LEFT JOIN (
  SELECT SUM(bloat_size) AS bloat_size,
         schemaname,
         tblname
  FROM (
    """ + INDEX_BTREE_BLOAT_SQL + """
  ) AS a
  GROUP BY schemaname, tblname
) AS ibloat
ON ibloat.schemaname = table_schema AND ibloat.tblname = table_name
JOIN pg_stat_all_tables
ON relname = table_name
WHERE table_schema = '{schema}'
AND table_name = '{table}';
    """
    return dict(conn.queryone(query.format(schema=schema, table=table)))


def check_table_exists(conn, schema, table):
    # Check that the specified table exists in schema
    rows = conn.query(
        "SELECT 1 FROM pg_tables WHERE tablename = '{table}' AND "
        "schemaname = '{schema}'".format(table=table, schema=schema)
    )
    if not list(rows):
        raise UserError("Table {}.{} not found".format(schema, table))


def check_index_exists(conn, schema, index):
    # Check that the specified table exists in schema
    rows = conn.query(
        "SELECT 1 FROM pg_indexes WHERE indexname = '{index}' AND "
        "schemaname = '{schema}'".format(index=index, schema=schema)
    )
    if not list(rows):
        raise UserError("Index {}.{} not found".format(schema, index))


def schedule_operation(operation_type, conn, database,
                       datetimeutc, app, table=None, index=None,
                       schema=None, **kwargs):
    # Schedule a maintenance operation (vacuum or analyze) statement through
    # background worker

    if table:
        check_table_exists(conn, schema, table)
    if index:
        check_index_exists(conn, schema, index)

    # Schedule a new task to background worker

    # We need to build a uniq id for this task to avoid scheduling twice the
    # same statement.
    m = hashlib.md5()
    m.update("{database}:{schema}:{table}{index}:{datetime}:{operation_type}"
             .format(database=database,
                     schema=schema or '',
                     table=table or '',
                     index=index or '',
                     datetime=datetimeutc,
                     operation_type=operation_type).encode('utf-8'))
    # Task scheduling
    try:
        # Convert string datetime to datetime object
        dt = datetime.strptime(datetimeutc, '%Y-%m-%dT%H:%M:%SZ')

        options = {
            'dbname': database,
        }
        if schema:
            options['schema'] = schema
        if table:
            options['table'] = table
        if index:
            options['index'] = index
        if 'mode' in kwargs:
            options['mode'] = kwargs['mode']

        res = app.scheduler.schedule_task(
            operation_type + '_worker',
            id=m.hexdigest()[:8],
            options=options,
            # We add one microsecond here to be compliant with scheduler
            # datetime format expected during task recovery
            start=(dt + timedelta(microseconds=1)),
            expire=0,
        )
    except Exception as e:
        logger.exception(str(e))
        raise HTTPError(500, "Unable to schedule %s" % operation_type)

    if res.type == taskmanager.MSG_TYPE_ERROR:
        logger.error(res.content)
        raise HTTPError(500, "Unable to schedule %s" % operation_type)

    return res.content


def schedule_vacuum(conn, database, mode, datetimeutc, app,
                    schema=None, table=None):
    return schedule_operation('vacuum', conn, database, datetimeutc, app,
                              mode=mode, schema=schema, table=table)


def vacuum(conn, dbname, mode, schema=None, table=None):
    # Run vacuum statement

    if table:
        check_table_exists(conn, schema, table)

    # Build the SQL query
    q = "VACUUM"
    q += " (%s) " % mode.upper() if mode else ""

    if schema and table:
        q += " {schema}.{table}".format(schema=schema, table=table)

    try:
        # Try to execute the statement
        logger.info("Running SQL on DB {}: {}".format(dbname, q))
        conn.execute(q)
        logger.info("VACCUM done.")
    except Exception as e:
        logger.exception(str(e))
        logger.error("Unable to execute SQL: %s" % q)
        message = "Unable to run vacuum %s" % mode
        if schema and table:
            message += " on {}.{}".format(schema, table)

        raise UserError(message)


def task_status_label(status):
    labels = ['todo', 'scheduled', 'queued', 'doing', 'done', 'failed',
              'canceled', 'aborted', 'abort']
    p = status.bit_length() - 1
    try:
        return labels[p]
    except IndexError:
        return 'unknown'


def list_scheduled_operation(app, operation_type, **kwargs):
    # Get list of scheduled vacuum operations
    ret = []
    try:
        # Ask it to the task manager
        tasks = taskmanager.TaskManager.send_message(
            str(os.path.join(app.config.temboard.home, '.tm.socket')),
            taskmanager.Message(taskmanager.MSG_TYPE_TASK_LIST, ''),
            authkey=None,
        )
    except Exception as e:
        logger.exception(str(e))
        raise HTTPError(500, "Unable to get scheduled vacuum list")

    for task in tasks:

        # We only want tasks for the operation type ('vacuum', 'analyze',
        # 'reindex')
        if task.worker_name != operation_type + '_worker':
            continue

        options = task.options
        # Filter by db/schema/table if provided
        if (all(k in kwargs for k in ['dbname', 'schema']) and
            (kwargs.get('dbname') != options.get('dbname') or
             kwargs.get('schema') != options.get('schema'))):
            continue

        if ('table' in kwargs and kwargs.get('table') != options.get('table')):
            continue
        if ('index' in kwargs and kwargs.get('index') != options.get('index')):
            continue

        ret.append(dict(
            id=task.id,
            dbname=options.get('dbname'),
            schema=options.get('schema'),
            table=options.get('table'),
            index=options.get('index'),
            mode=options.get('mode'),
            datetime=task.start_datetime.strftime("%Y-%m-%dT%H:%M:%SZ"),
            status=task_status_label(task.status)
        ))
    return ret


def list_scheduled_vacuum(app, **kwargs):
    return list_scheduled_operation(app, 'vacuum', **kwargs)


def schedule_analyze(conn, database, datetimeutc, app,
                     schema=None, table=None):
    return schedule_operation('analyze', conn, database, datetimeutc, app,
                              schema=schema, table=table)


def analyze(conn, dbname, schema=None, table=None):
    # Run analyze statement
    if table:
        check_table_exists(conn, schema, table)

    # Build the SQL query
    q = "ANALYZE"
    if schema and table:
        q += " {schema}.{table}".format(schema=schema, table=table)

    try:
        # Try to execute the statement
        logger.info("Running SQL on DB {}: {}".format(dbname, q))
        conn.execute(q)
        logger.info("ANALYZE done.")
    except Exception as e:
        logger.exception(str(e))
        logger.error("Unable to execute SQL: %s" % q)
        message = "Unable to run analyze"
        if schema and table:
            message += " on {}.{}".format(schema, table)

        raise UserError(message)


def list_scheduled_analyze(app, **kwargs):
    return list_scheduled_operation(app, 'analyze', **kwargs)


def schedule_reindex(conn, database, datetimeutc, app,
                     schema=None, table=None, index=None):
    return schedule_operation('reindex', conn, database, datetimeutc, app,
                              schema=schema, table=table, index=index)


def reindex(conn, dbname, schema, table, index):
    if index:
        check_index_exists(conn, schema, index)
    if table:
        check_table_exists(conn, schema, table)

    # Build the SQL query
    q = "REINDEX"
    if table:
        element = '{schema}.{table}'.format(schema=schema, table=table)
        q += " TABLE {element}".format(element=element)
    elif index:
        element = '{schema}.{index}'.format(schema=schema, index=index)
        q += " INDEX {element}".format(element=element)
    else:
        element = '{dbname}'.format(dbname=dbname)
        q += " DATABASE {element}".format(element=element)

    try:
        # Try to execute the statement
        logger.info("Running SQL on DB {}: {}".format(dbname, q))
        conn.execute(q)
        logger.info("reindex done.")
    except Exception as e:
        logger.exception(str(e))
        logger.error("Unable to execute SQL: %s" % q)
        raise UserError("Unable to run reindex on %s" % (element))


def list_scheduled_reindex(app, **kwargs):
    return list_scheduled_operation(app, 'reindex', **kwargs)


def cancel_scheduled_operation(id, app):
    # Cancel one scheduled operation. If the operation is running, the task
    # is going to be aborted.

    # Check the id
    if id not in [t['id'] for t in
                  list_scheduled_vacuum(app) +
                  list_scheduled_analyze(app) +
                  list_scheduled_reindex(app)]:
        raise HTTPError(404, "Scheduled operation not found")

    try:
        # Ask it to the task manager
        taskmanager.TaskManager.send_message(
            str(os.path.join(app.config.temboard.home, '.tm.socket')),
            taskmanager.Message(
                taskmanager.MSG_TYPE_TASK_CANCEL,
                dict(task_id=id),
            ),
            authkey=None,
        )
    except Exception as e:
        logger.exception(str(e))
        raise HTTPError(500, "Unable to cancel operation")

    return dict(response="ok")



# The configuration API merges args, file, environment and defaults safely,
# even when reloading.

import logging
import os
from argparse import SUPPRESS as SUPPRESS_ARG

from .utils import DotDict
from .errors import UserError


logger = logging.getLogger(__name__)


class OptionSpec(object):
    REQUIRED = object()

    # Hold known name and default of an option.
    #
    # An option *must* be specified to follow the principle of *validated your
    # inputs*.
    #
    # Defining defaults here is agnostic from origin : argparse, environ,
    # ConfigParser, etc. The origin of configuration must not take care of
    # default nor validation.
    #
    # Set default to OptionSpec.REQUIRED to enforce user definition of option
    # value.
    #
    # Option value can be None, meaning it is undefined.

    def __init__(self, section, name, validator=None, default=None):
        self.section = section
        self.name = name
        self.default = default
        self.validator = validator

    def __repr__(self):
        return '<OptionSpec %s>' % (self,)

    def __str__(self):
        return '%s_%s' % (self.section, self.name)

    def __eq__(self, other):
        return str(self) == str(other)

    def __hash__(self):
        return hash(str(self))

    @property
    def required(self):
        return self.default is self.REQUIRED

    def add_argument(self, parser, *a, **kw):
        help_ = kw.get('help')
        if help_:
            help_ = help_ % dict(
                default="*required*" if self.required else self.default,
            )
            # Escape % for Argparse formatting.
            kw['help'] = help_.replace("%", "%%")

        kw.setdefault('dest', str(self))
        kw.setdefault('default', SUPPRESS_ARG)

        return parser.add_argument(*a, **kw)

    def validate(self, value):
        if value.value is None or not self.validator:
            return value.value

        try:
            return self.validator(value.value)
        except ValueError as e:
            logger.error(
                "Invalid %s from %s: %s...", value.name, value.origin, e)
            raise


class Value(object):
    # Hold an option value and its origin
    def __init__(self, name, value, origin):
        self.name = name
        self.value = value
        self.origin = origin

    def __repr__(self):
        return '<%(name)s=%(value)r %(origin)s>' % self.__dict__


def iter_args_values(args):
    # Walk args from argparse and yield values.
    if not args:
        return

    for k, v in args.__dict__.items():
        yield Value(k, v, 'args')


def iter_configparser_values(parser, filename='config'):
    for section in parser.sections():
        for name, value in parser.items(section):
            name = '%s_%s' % (section, name)

            if value.startswith('"') and value.startswith('"'):
                raise UserError(
                    "You must not quote string in configuration (%s)."
                    % name.replace('_', '.')
                )

            if value.startswith("'") and value.startswith("'"):
                raise UserError(
                    "You must not quote string in configuration (%s)."
                    % name.replace('_', '.')
                )

            yield Value(name, value, origin=filename)


def iter_environ_values(environ):
    environ = environ or {}
    prefix = 'TEMBOARD_'
    for k, v in environ.items():
        if not k.startswith(prefix):
            continue

        k = k.lower()
        if hasattr(v, 'decode'):
            v = v.decode('utf-8')

        # Yield the value with temboard prefix so we don't have to define
        # TEMBOARD_TEMBOARD_* to set a value in temboard section.
        yield Value(k, v, 'environ')
        yield Value(k[len(prefix):], v, 'environ')


def iter_defaults(specs):
    # Walk specs flat dict and yield default values.
    for spec in specs.values():
        if spec.required:
            continue
        yield Value(str(spec), spec.default, 'defaults')


class MergedConfiguration(DotDict):
    # Merge and holds configuration from args, files and more
    #
    # Origin order: args > environ > file > defaults

    def __init__(self, specs=None):
        DotDict.__init__(self)  # PY2, move to super
        self.__dict__['specs'] = dict([(s, s) for s in specs or []])
        self.__dict__['unvalidated_specs'] = set(self.specs)

    def add_specs(self, specs):
        for s in specs:
            self.specs[s] = s
            self.unvalidated_specs.add(s)

    def remove_specs(self, specs):
        self.unvalidated_specs -= set(specs)
        for s in specs:
            self.specs.pop(s, None)
            # Clean value
            try:
                del self[s.section][s.name]
            except KeyError:
                pass

            # Clean section if empty
            if not self.get(s.section):
                self.pop(s.section, None)

    def add_values(self, values):
        # Search missing values in values and validate them.

        values = dict((v.name, v) for v in values)
        for name in list(self.unvalidated_specs):
            try:
                value = values[name]
            except KeyError:
                continue

            spec = self.specs[name]
            value = spec.validate(value)
            section = self.setdefault(spec.section, {})
            section[spec.name] = value
            self.unvalidated_specs.remove(name)

    def check_required(self):
        for name in self.unvalidated_specs:
            spec = self.specs[name]
            if spec.required:
                msg = "Missing %s:%s configuration" % (spec.section, spec.name)
                raise UserError(msg)

    def load(self, args=None, environ=None, parser=None, pwd=None,
             reload_=False):
        # Origins are loaded in order. First wins..
        #
        # Loading in this order avoid validating ignored values.

        if reload_:
            # Reset unvalidated set to re-analyze all options.
            self.unvalidated_specs.update(set(self.specs))

        try:
            self.add_values(iter_args_values(args))
            self.add_values(iter_environ_values(environ))
        except ValueError as e:
            logger.debug("Bad value %s.", e)
            raise UserError("Failed to load configuration.")

        if pwd:
            try:
                # configfile and default values are relative to configfile
                # directory.
                oldpwd = os.getcwd()
            except OSError as e:
                raise UserError("Failed to get CWD: %s" % e)

            try:
                os.chdir(pwd)
            except OSError as e:
                raise UserError("Failed to access %s: %s" % (pwd, e))
        else:
            oldpwd = None

        try:
            if parser:
                self.add_values(iter_configparser_values(parser))
            self.add_values(iter_defaults(self.specs))
        except ValueError as e:
            logger.debug("Bad value %s.", e)
            raise UserError("Failed to load configuration.")

        if oldpwd:
            try:
                os.chdir(oldpwd)
            except OSError as e:
                logger.warning("Can't move back to %s: %s", oldpwd, e)

        self.check_required()



import re

from bottle import Bottle, default_app, HTTPError, request

from . import functions as pgconf_functions


bottle = Bottle()


@bottle.get('/configuration')
def get_configuration(pgconn):
    return get_configuration_category(pgconn, None)


@bottle.get('/configuration/category/<category:path>')
def get_configuration_category(pgconn, category):
    search = None
    if 'filter' in request.query:
        if not re.match('([a-zA-Z0-9_]{3,128})', request.query['filter']):
            raise HTTPError(406, "Parameter 'filter' is malformed")
        search = request.query['filter']
    if category:
        # Unquote +.
        category = category.replace('+', ' ')
    return pgconf_functions.get_settings(pgconn, category, search)


@bottle.get('/configuration/categories')
def get_configuration_categories(pgconn):
    return pgconf_functions.get_settings_categories(pgconn)


@bottle.post('/configuration')
def post_configuration(pgconn):
    if 'settings' not in request.json:
        raise HTTPError(406, "Parameter 'settings' not sent.")
    current = get_configuration_category(pgconn, None)
    return pgconf_functions.post_settings(
        default_app().temboard, pgconn,
        current, request.json['settings'],
    )


@bottle.get('/configuration/status')
def get_status(pgconn):
    return pgconf_functions.get_settings_status(pgconn)


class PgConfPlugin:
    PG_MIN_VERSION = (90500, 9.5)

    def __init__(self, app, **kw):
        self.app = app

    def load(self):
        default_app().mount('/pgconf/', bottle)



# Introspect Postgres instance, temBoard agent and system
#
# Discover is a stable set of properties identifying a running system. temBoard
# computes an ETag from discover data to ease change detection.
#
# Discover data is cached in a discover.json file. temBoard refreshes this file
# in two cases : start of agent, connection lost.
#

import hashlib
import json
import logging
import os
import socket
import sys
import time
from datetime import datetime
from platform import machine, python_version
from multiprocessing import cpu_count

from .core import workers
from .queries import QUERIES
from .toolkit.errors import UserError
from .toolkit.versions import (
    format_pq_version,
    read_distinfo,
    read_libpq_version,
)
from .version import __version__


logger = logging.getLogger(__name__)


class Discover:
    def __init__(self, app):
        self.app = app
        self.path = self.app.config.temboard.home + '/discover.json'
        self.data = dict(postgres={}, system={}, temboard={})
        self.json = None  # bytes
        self.etag = None
        self.file_etag = None
        self.mtime = None
        self.inhibit_observer = False

    def connection_lost(self):
        if self.inhibit_observer:
            return
        # Callback for postgres.ConnectionPool connection lost event.
        logger.info("Queueing discover refresh.")
        discover.defer(self.app)

    def ensure_latest(self):
        if self.mtime != os.stat(self.path).st_mtime:
            logger.debug("Discover file changed.")
            return self.read()
        return self.data

    def read(self):
        logger.debug("Reading discover data from %s.", self.path)
        try:
            fo = open(self.path, 'rb')
        except IOError as e:
            logger.debug("Failed to read manifest: %s.", e)
            return self.data

        with fo:
            self.json = fo.read()

        self.etag = self.file_etag = compute_etag(self.json)

        try:
            data = json.loads(self.json.decode('utf-8'))
        except json.JSONDecodeError as e:
            raise UserError("Malformed manifest: %s" % e)

        if not isinstance(data, dict):
            raise UserError("Malformed manifest: not a mapping")

        self.data.update(data)
        self.mtime = os.stat(self.path).st_mtime
        return self.data

    def write(self, fo=None):
        if fo is None:
            if self.etag == self.file_etag:
                logger.debug("Discover file up to date.")
                return
            self.mtime = None

        with (fo or open(self.path, 'w')) as fo:
            fo.write(self.json.decode('utf-8'))

        if self.mtime is None:  # if not sys.stdout.
            logger.debug("Wrote discover.json with ETag %s.", self.etag)
            self.mtime = os.stat(self.path).st_mtime
            self.file_etag = self.etag

    def refresh(self, conn=None):
        logger.debug("Inspecting temBoard and system.")
        d = self.data
        old_postgres = self.data.get('postgres', {})
        d.clear()

        d['postgres'] = {}
        d['system'] = {}
        d['temboard'] = {}

        d['temboard']['bin'] = sys.argv[0]
        d['temboard']['configfile'] = self.app.config.temboard.configfile
        d['temboard']['plugins'] = self.app.config.temboard.plugins

        d['system']['fqdn'] = self.app.config.temboard.hostname
        collect_versions(d)
        collect_cpu(d)
        collect_memory(d)
        collect_system(d)

        try:
            with self.app.postgres.maybe_connect(conn) as pgconn:
                logger.debug("Inspecting Postgres instance.")
                collect_postgres(d, pgconn)
        except Exception as e:
            logger.error("Failed to collect Postgres data: %s", e)
            d['postgres'] = old_postgres

        # Build JSON to compute ETag.
        json_text = json.dumps(
            self.data,
            indent="  ",
            sort_keys=True,
        ) + "\n"
        self.json = json_text.encode('utf-8')
        self.etag = compute_etag(self.json)

        if self.etag != self.file_etag:
            logger.info("Instance discover updated.")
        else:
            logger.debug("Instance discover has not changed.")

        return self.data


def collect_cpu(data):
    s = data['system']
    s['cpu_count'] = cpu_count()

    with open('/proc/cpuinfo') as fo:
        for line in fo:
            if not line.startswith('model name\t'):
                continue
            _, _, model = line.partition("\t: ")
            s['cpu_model'] = model.rstrip()


def collect_memory(data):
    meminfo = {}
    with open('/proc/meminfo', 'r') as fo:
        for line in fo:
            if 'kB' not in line:
                continue
            field, value, kb = line.split()
            meminfo[field[:-1]] = int(value) * 1024

    s = data['system']
    s['memory'] = meminfo['MemTotal']
    s['swap'] = meminfo['SwapTotal']
    s['hugepage'] = meminfo['Hugepagesize']


def collect_postgres(data, conn):
    row = conn.queryone(QUERIES['discover'])
    data['postgres'].update(row)
    data['postgres']['start_time'] = row['start_time'].strftime(
        "%Y-%m-%dT%H:%M:%S%Z"
    )

    for row in conn.query(QUERIES['discover-settings']):
        t = row['vartype']
        v = row['setting']
        if 'integer' == t:
            v = int(v)
        elif 'bool' == t:
            v = 'on' == v

        u = row['unit']
        if u is None or 'B' == u:
            pass
        elif '8kB' == u:
            v = v * 8 * 1024
        else:
            raise ValueError("Unsupported unit %s" % u)
        data['postgres'][row['name']] = v
    pid_file_path = find_pid_file(data['postgres'])
    data['postgres']['pid'] = get_postmaster_pid(pid_file_path)


def get_postmaster_pid(pid_file):
    """Return the pid of the postmaster process"""
    with open(pid_file) as f:
        return int(f.readline().split()[0])


def find_pid_file(postgres_data):
    """Return the pid file path"""
    if postgres_data["external_pid_file"]:
        return postgres_data["external_pid_file"]
    return '%s/postmaster.pid' % postgres_data["data_directory"]


def collect_system(data):
    uname = os.uname()
    s = data['system']
    s['os'] = uname.sysname
    s['os_version'] = uname.release
    s['arch'] = machine()
    s['hostname'] = socket.gethostname()
    with open('/proc/uptime', 'r') as f:
        uptime_seconds = float(f.readline().split()[0])
        s['start_time'] = datetime.utcfromtimestamp(
            int(time.time() - uptime_seconds)
        ).strftime("%Y-%m-%dT%H:%M:%S%Z")


def collect_versions(data):
    versions = inspect_versions()
    data['temboard'].update(dict(
        agent_version=versions['temboard'],
    ))

    for k in 'bottle', 'cryptography', 'libpq', 'psycopg2', 'python':
        data['temboard'][k + '_version'] = versions[k]

    data['temboard']['pythonbin'] = versions['pythonbin']
    dist = versions['distname'] + ' ' + versions['distversion']
    data['system']['distribution'] = dist


def inspect_versions():
    from bottle import __version__ as bottle_version
    from psycopg2 import __version__ as psycopg2_version
    from cryptography import __version__ as cryptography_version

    distinfos = read_distinfo()

    return dict(
        temboard=__version__,
        temboardbin=sys.argv[0],
        psycopg2=psycopg2_version,
        python=python_version(),
        pythonbin=sys.executable,
        bottle=bottle_version,
        distname=distinfos['NAME'],
        distversion=distinfos.get('VERSION', 'n/a'),
        libpq=format_pq_version(read_libpq_version()),
        cryptography=cryptography_version,
    )


def compute_etag(data):
    h = hashlib.new('sha256')
    h.update(data)
    return h.hexdigest()


@workers.register(pool_size=1)
def discover(app):
    """ Refresh discover data. """
    app.discover.ensure_latest()
    app.discover.inhibit_observer = True
    app.discover.refresh()
    app.discover.inhibit_observer = False
    app.discover.write()
